!*****************************************************************************************
!MODULE CLPS----------------------------------------------------------- BY SML. 14/08/2012
!*****************************************************************************************

 MODULE CLPS

!*****************************************************************************************
!CONTAINS SUBROUTINES "COLLAPSE" & "RETRIEVE", USED -RESPECTIVELY- TO COLLAPSE A SYMMETRIC
!SYSTEM OF EQUATIONS STORED IN SPARSE IJA FORM WITH ZERO ROWS AND TO RETRIEVE THE ORIGINAL
!STRUCTURE OF THE SOLUTION'S VECTOR.
!*****************************************************************************************
!
!***VERSION 1.1***
!AUXILIARY MODULE "PARAM" WITH NUMERICAL PRECISION DEFINITION IS REQUIRED.
!***
!
!-------
!MODULES
!-------
 USE PARAM
!---
!
 CONTAINS

!*****************************************************************************************
!SUBROUTINE COLLAPSE--------------------------------------------------- BY SML. 01/12/2008
!*****************************************************************************************
!
!COLLAPSE A SYMMETRIC SYSTEM OF EQUATIONS WITH ZERO ROWS STORED IN SPARSE IJA FORM.
!***
!
!***VERSION 1.0***

 SUBROUTINE COLLAPSE (N, IA, JA, R, ZROW, DCOL)

!INTENT(INOUT):    N - DIMENSION OF THE SYSTEM.
!INTENT(INOUT):   IA - INTEGER VECTOR WITH EACH ROW FIRST INDEX.
!INTENT(INOUT):   JA - INTEGER VECTOR WITH THE COLUMN NUMBER OF EACH ENTRY.
!INTENT(INOUT):    R - RIGHT HAND SIDE OF THE SYSTEM
!INTENT(IN)   : ZROW - ZERO ROW INDICATOR BINARY VECTOR (0,1): 0=ZERO ROW; 1=NONZERO ROW. 
!INTENT(IN)   : DCOL - COUNTS THE NUMBER OF POSITIONS COLUMN INDEXES SHOULD BE MOVED.
!***

!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE

!DUMMY ARGUMENTS
 INTEGER(K4B), INTENT(INOUT)::  N
 INTEGER(K4B), INTENT(INOUT):: IA(:)
 INTEGER(K4B), INTENT(INOUT):: JA(:)

 REAL(DP), INTENT(INOUT):: R(:)

 INTEGER(K4B), INTENT(IN):: ZROW(:)
 INTEGER(K4B), INTENT(IN):: DCOL(:)
!***

!LOCAL VARIABLES 
 INTEGER(K4B):: DROW !DROW: COUNTS THE NUMBER OF DELETED ZERO ROWS.
 INTEGER(K4B)::  NZE	! NZE: COUNTS NON ZERO EQUATIONS.
 INTEGER(K4B):: i, j !DUMMY VARIABLES.
!***

!*********
!PROCEDURE
!*********

!COLLAPSE ROWS!
 DROW = 0
  NZE = 0

 DO i = 1, N

	!IF i IS A NONZERO ROW:
	 IF (ZROW(i).EQ.0) THEN

		 NZE = NZE + 1 

 	!IF i IS A ZERO ROW:
	 ELSE
			
		 DROW = DROW + 1
		 CYCLE

	 ENDIF

	 IA(NZE) = IA(i)
	  R(NZE) =  R(i)

	!COLLAPSE COLUMNS
	 DO j = IA(i), IA(i+1) - 1
	 JA(j) =  JA(j) - DCOL(JA(j))
	 ENDDO

 ENDDO

 IA((N-DROW)+1) = IA(N + 1) !SAVE POSITION (NEQ+1) OF IA VECTOR
 N = N - DROW !COLLAPSE THE SYSTEM
!***

 END SUBROUTINE	COLLAPSE
!***

!*****************************************************************************************
!SUBROUTINE RETRIEVE--------------------------------------------------- BY SML. 02/12/2008
!*****************************************************************************************
!
!RETRIEVE THE SOLUTION'S VECTOR ORIGINAL STRUCTURE OF A SYMMETRIC SYSTEM OF EQUATIONS WHE- 
!RE SOME ROWS HAVE BEEN DELETED, AND SET THOSE ROW'S SOLUTIONS TO A DEFINED VALUE.
!
!***VERSION 1.0***

 SUBROUTINE RETRIEVE (N, ZROW, SOL, VAL)

!INTENT(IN)   :    N - DIMENSION OF THE REDUCED SYSTEM.
!INTENT(IN)   : ZROW - ZERO ROW INDICATOR BINARY VECTOR (0,1): 0=ZERO ROW; 1=NONZERO ROW. 
!INTENT(IN)   :  VAL - DEFINED VALUE FOR THE DELETED ROWS SOLUTIONS.
!INTENT(INOUT):  SOL - SOLUTIONS OF THE REDUCED SYSTEM.
!***

!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE

!DUMMY ARGUMENTS
 INTEGER(K4B), INTENT(IN):: N
 INTEGER(K4B), INTENT(IN):: ZROW(:)
 REAL(DP), INTENT(IN):: VAL

 REAL(DP), INTENT(INOUT):: SOL(:)
!***

!LOCAL VARIABLES 
 INTEGER(K4B):: OLDN !ORIGINAL EXTENT OF THE SYSTEM.
 INTEGER(K4B):: i, CONT !DUMMY VARIABLES.

 REAL(DP):: TEMP(N) 
!***

!*********
!PROCEDURE
!*********

!RETRIEVE THE ORIGINAL EXTENT:
 OLDN = SIZE(SOL)

 TEMP = SOL(1:N)
 CONT = 0

 DO i = 1, OLDN

	!IF i IS A NONZERO ROW:
	 IF (ZROW(i).EQ.0) THEN

		 CONT = CONT + 1
 		 SOL(i) = TEMP(CONT)

	!IF i IS A ZERO ROW:
	 ELSE

		 SOL(i) = VAL

	 ENDIF

 ENDDO
!***

 END SUBROUTINE	RETRIEVE
!***


 END MODULE CLPS
!***