!********************************************************************************
!MODULE ITPACK------------------------------------------------ BY SML. 14/08/2012
!********************************************************************************

 MODULE ITPACK

!********************************************************************************
!THIS MODULE CONTAINS SUBROUTINES "SBINI", "SBSIJ" & "SBEND" (AND AUXILLIARY SUB-
!ROUTINES "IVFILL" AND "VFILL") FROM THE PACKAGE "ITPACK" DEVELOPED BY THE CENTER
!FOR NUMERICAL ANALYSIS, UNIVERSITY OF TEXAS. PERFORM SPARSE MATRIX STORAGE UNDER
!IJA STRUCTURE. IT CONTAINS A LINKED LIST SERACHING ALGORITHM, THAT ALLOW SEVERAL
!CONTRIBUTIONS TO THE SAME POSITION IN THE ORIGINAL MATRIX.
!********************************************************************************
!
!***VERSION 1.1***
!AUXILIARY MODULE "PARAM" WITH NUMERICAL PRECISION DEFINITION IS REQUIRED.
!***
!
!-------
!MODULES
!-------
 USE PARAM
!---

 CONTAINS

!***********************************************************************!ITP00010
 SUBROUTINE SBEND ( N, NZ, IA, JA, A, IWORK )                           !ITP00020
!***********************************************************************!ITP00030
!                                                                       !ITP00040
!SBEND IS THE THIRD OF A SUITE OF SUBROUTINES TO AID THE				!ITP00050
!USER TO CONSTRUCT THE  IA, JA, A DATA STRUCTURE USED IN				!ITP00060
!ITPACK.																!ITP00070
!                                                                       !ITP00080
!SBEND RESTRUCTURES THE LINKED LIST DATA STRUCTURE BUILT BY				!ITP00090
!SBINI AND SBSIJ INTO THE FINAL DATA STRUCTURE REQUIRE BY				!ITP00100
!ITPACK. THE RESTRUCTURING CAN TAKE PLACE IN THE MINIMUM				!ITP00110
!AMOUNT OF MEMORY REQUIRED TO HOLD THE NONZERO STRUCTURE OF				!ITP00120
!THE SPARSE MATRIX BUT WILL RUN QUICKER IF MORE STORAGE					!ITP00130
!IS ALLOWED.															!ITP00140
!                                                                       !ITP00150
!SBEND IS BASED ON SUBROUTINE BUILD OF THE SPARSE MATRIX				!ITP00160
!PACKAGE SPARSPAK DEVELOPED BY ALAN GEORGE AND JOSEPH LUI				!ITP00170
!OF THE UNIVERSITY OF WATERLOO, WATERLOO, ONTARIO.						!ITP00180
!                                                                       !ITP00190
!PARAMETERS:	                                                        !ITP00200
!			                                                            !ITP00210
!INPUT			                                                        !ITP00220
!                                                                       !ITP00230
!	  N       THE ORDER OF THE LINEAR SYSTEM                            !ITP00240
!                                                                       !ITP00250
!     NZ      THE LENGTH OF THE ARRAYS JA, IWORK, AND A.                !ITP00260
!                                                                       !ITP00270
!INPUT/OUTPUT															!ITP00280
!                                                                       !ITP00290
!     IA      INTEGER ARRAY OF LENGTH N+1. THE FIRST N ENTRIES			!ITP00300
!             POINT TO THE BEGINNING OF THE LINKED LIST FOR EACH        !ITP00310
!             ROW. [IA(N+1)+1] IS THE TOP OF THE LINKED LISTS			!ITP00320
!             CONTAINED IN JA, IWORK, AND A. ON OUTPUT IA WILL			!ITP00330
!             POINT TO THE FIRST ENTRY OF EACH ROW IN THE FINAL         !ITP00340
!             DATA STRUCTURE.                                           !ITP00350
!                                                                       !ITP00360
!     JA      INTEGER ARRAY OF LENGTH NZ. ON INPUT JA STORES THE		!ITP00370
!             COLUMN NUMBERS OF THE NONZERO ENTRIES AS INDICATED        !ITP00380
!             BY THE LINKED LISTS. ON OUTPUT JA STORES THE				!ITP00390
!             COLUMN NUMBERS IN ROW ORDERED FORM.                       !ITP00400
!                                                                       !ITP00410
!     A       REAL ARRAY OF LENGTH NZ. ON INPUT A STORES THE			!ITP00420
!             VALUE OF THE NOZERO ENTRIES AS INDICATED BY THE           !ITP00430
!             LINKED LISTS. ON OUTPUT A STORES THE VALUES IN			!ITP00440
!             ROW ORDERED FORM.                                         !ITP00450
!                                                                       !ITP00460
!     IWORK   INTEGER ARRAY OF LENGTH NZ. ON INPUT IWORK STORES			!ITP00470
!             THE LINKS OF THE LINKED LISTS. ON OUTPUT IT IS			!ITP00480
!             DESTROYED.                                                !ITP00490
!                                                                       !ITP00500
!***********************************************************************!ITP00510
 INTEGER(K4B), INTENT(IN):: N, NZ										!ITP00520
 INTEGER(K4B), INTENT(INOUT):: IA(:), JA(:), IWORK(:)					!ITP00530
 REAL(DP), INTENT(INOUT):: A(:)									        !ITP00540
!                                                                       !ITP00550
 INTEGER(K4B):: MAXTOP, NEXT, TOP, IDEG, NULINK, JAJ, HLINK				!ITP00560
 INTEGER(K4B)::	OHLINK, L, I, LINK, MHLINK								!ITP00570
 REAL(DP):: VAL												        	!ITP00580
!                                                                       !ITP00590
!***********************************************************************!ITP00600
!                                                                       !ITP00610
!INITIALIZATION															!ITP00620
!                                                                       !ITP00630
!THE VARIABLES NEXT AND TOP RESPECTIVELY POINT TO THE					!ITP00640
!NEXT AVAILABLE ENTRY FOR THE FINAL DATA STRUCTURE AND					!ITP00650
!THE TOP OF THE REMAINDER OF THE LINKED LISTS.							!ITP00660
!                                                                       !ITP00670
   NEXT = 1																!ITP00680
    TOP = IA(N+1) + 1													!ITP00690
 MAXTOP = NZ - IA(N+1) + 1												!ITP00700
!                                                                       !ITP00710
!***********************************************************************!ITP00720
!                                                                       !ITP00730
!CONVERT EACH ROW INTO FINAL FORM										!ITP00740
!                                                                       !ITP00750
    DO 90 I = 1, N														!ITP00760
		IDEG = 0														!ITP00770
		NULINK = IA(I)													!ITP00780
!                                                                       !ITP00790
!	    LOOP OVER EACH NODE IN THE LINKED LIST OF ROW I					!ITP00800
!                                                                       !ITP00810
 10     LINK = NULINK													!ITP00820
        IF ( LINK .LE. 0 ) GOTO 80										!ITP00830
        NULINK = IWORK(LINK)											!ITP00840
        JAJ    = JA(LINK)												!ITP00850
        VAL    = A(LINK)												!ITP00860
!                                                                       !ITP00870
!		CHECK TO SEE IF A COLLISION BETWEEN THE LINKED LISTS            !ITP00880
!		AND THE FINAL FORM HAS OCCURRED.                                !ITP00890
!                                                                       !ITP00900
        IF ( NEXT .GE. TOP  .AND.  LINK .NE. TOP ) GOTO 20				!ITP00910
!                                                                       !ITP00920
!		COLLISION HAS NOT OCCURRED. FREE THE SPACE FOR THE TRIPLE       !ITP00930
!		(JA(LINK), A(LINK), IWORK(LINK))                                !ITP00940
!                                                                       !ITP00950
        JA(LINK)    = 0													!ITP00960
        A(LINK)     = 0.D0												!ITP00970
        IWORK(LINK) = 0													!ITP00980
!                                                                       !ITP00990
!		SPECIAL CASE TO MOVE TOP DOWN IF LINK .EQ. TOP					!ITP01000
        IF ( LINK .EQ. TOP ) GOTO 60                                    !ITP01010
!																		!ITP01020
        GOTO 70															!ITP01030
!                                                                       !ITP01040
!***********************************************************************!ITP01050
!                                                                       !ITP01060
!		COLLISION HAS OCCURRED. CLEAR OFF SOME SPACE FOR THE CURRENT	!ITP01070
!		ENTRY BY MOVING THE TRIPLE ( JA(TOP),A(TOP),IWORK(TOP) )		!ITP01080
!		DOWNWARDS TO THE FREED TRIPLE ( JA(LINK),A(LINK),IWORK(LINK) ).	!ITP01090
!		THEN ADJUST THE LINK FIELDS.									!ITP01100
!                                                                       !ITP01110
!		PATCH UP THE LINKED LIST FOR THE CURRENT ROW I. THEN			!ITP01120
!		TRAVERSE THE LINKED LIST CONTAINING TOP UNTIL THE				!ITP01130
!		POINTER BACK TO IA IS FOUND.									!ITP01140
!                                                                       !ITP01150
 20     IA(I) = LINK													!ITP01160
        HLINK = TOP														!ITP01170
!                                                                       !ITP01180
 30     HLINK = IWORK(HLINK)											!ITP01190
        IF ( HLINK .GT. 0 ) GOTO 30										!ITP01200
!                                                                       !ITP01210
!		NOW FOLLOW THE LINKED LIST BACK TO TOP KEEPING TRACK			!ITP01220
!		OF THE OLD LINK.												!ITP01230
        MHLINK = -HLINK													!ITP01240
																		!ITP01250
		IF ( IA(MHLINK) .NE. TOP ) GOTO 40								!ITP01260
!																		!ITP01270
!       SPECIAL CASE IF IA(-HLINK) = TOP								!ITP01280
!		(FOUND TOP POSITION FOR THE LINKED LIST)                        !ITP01290
        IWORK(LINK) = IWORK(TOP)										!ITP01300
        JA(LINK)    = JA(TOP)											!ITP01310
        A(LINK)     = A(TOP)											!ITP01320
        IA(MHLINK)  = LINK												!ITP01330
        IF ( NULINK .EQ. TOP ) NULINK = LINK							!ITP01340
        GOTO 60															!ITP01350
!                                                                       !ITP01360
!		USUAL CASE (IA(-HLINK) IS NOT TOP).								!ITP01370
!                                                                       !ITP01380
 40		HLINK  = IA(MHLINK)												!ITP01390
 50		OHLINK = HLINK													!ITP01400
        HLINK  = IWORK(OHLINK)											!ITP01410
        IF ( HLINK .NE. TOP ) GOTO 50									!ITP01420
!                                                                       !ITP01430
		IWORK(LINK)   = IWORK(TOP)										!ITP01440
		JA(LINK)     = JA(TOP)											!ITP01450
		A(LINK)      = A(TOP)											!ITP01460
        IF ( OHLINK .NE. LINK ) IWORK(OHLINK) = LINK					!ITP01470
        IF ( NULINK .EQ. TOP ) NULINK = LINK							!ITP01480
!                                                                       !ITP01490
!		COLLAPSE TOP OF LINK LIST BY AS MUCH AS POSSIBLE				!ITP01500
!                                                                       !ITP01510
 60		TOP = TOP + 1													!ITP01520
        IF ( TOP .GE. MAXTOP ) GOTO 70									!ITP01530
        IF ( IWORK(TOP) .NE. 0 ) GOTO 70								!ITP01540
        GOTO 60															!ITP01550
!                                                                       !ITP01560
!***********************************************************************!ITP01570
!                                                                       !ITP01580
!		PUT THE CURRENT TRIPLE INTO THE FINAL DATA STRUCTURE			!ITP01590
!                                                                       !ITP01600
 70		JA(NEXT) = JAJ													!ITP01610
        A(NEXT)  = VAL													!ITP01620
        NEXT     = NEXT + 1												!ITP01630
        IDEG     = IDEG + 1												!ITP01640
        GOTO 10															!ITP01650
!                                                                       !ITP01660
!		FINAL STRUCTURE FOR ROW I IS COMPLETE. LINKED LIST IS			!ITP01670
!		DESTROYED AND WILL BE RECAPTURED AS NECESSARY BY THE            !ITP01680
!		LOOP ON LABEL 60                                                !ITP01690
!                                                                       !ITP01700
 80		IA(I) = IDEG													!ITP01710
!                                                                       !ITP01720
 90 CONTINUE															!ITP01730
!                                                                       !ITP01740
!***********************************************************************!ITP01750
!                                                                       !ITP01760
!FINALIZE THE DATA STRUCTURE BY BUILDING THE FINAL VERSION OF IA.		!ITP01770
!                                                                       !ITP01780
	 L = IA(1) + 1														!ITP01790
     IA(1) = 1															!ITP01800
     DO 100 I = 1, N													!ITP01810
		IDEG = IA(I+1)													!ITP01820
        IA(I+1) = L														!ITP01830
        L = L + IDEG													!ITP01840
 100    CONTINUE														!ITP01850
!                                                                       !ITP01860
!FINAL IA, JA, A DATA STRUCTURE BUILT.									!ITP01870
																		!ITP01880
!		                                                                !ITP01890
 END SUBROUTINE SBEND													!ITP01900
!																		!ITP01910
!***********************************************************************!ITP01920
 SUBROUTINE SBINI ( N, NZ, IA, JA, A, IWORK )                           !ITP01930
!***********************************************************************!ITP01940
!                                                                       !ITP01950
!SBINI IS THE FIRST OF A SUITE OF THREE SUBROUTINES TO AID				!ITP01960
!THE USER TO CONSTRUCT THE IA, JA, A DATA STRUCTURE USED				!ITP01970
!IN ITPACK.																!ITP01980
!                                                                       !ITP01990
!SBINI INITIALIZES THE ARRAYS IA, JA, IWORK, AND A. THE OTHER			!ITP02000
!SUBROUTINES IN THE SUITE ARE SBSIJ (WHICH BUILDS A LINKED				!ITP02010
!LIST REPRESENTATION OF THE MATRIX STRUCTURE) AND SBEND (WHICH			!ITP02020
!RESTRUCTURE THE LINKED LIST FORM INTO THE FINAL FORM).					!ITP02030
!                                                                       !ITP02040
!PARAMETERS																!ITP02050
!                                                                       !ITP02060
!INPUT																	!ITP02070
!                                                                       !ITP02080
!      N         THE ORDER OF THE LINEAR SYSTEM                         !ITP02090
!                                                                       !ITP02100
!     NZ         THE MAXIMUM NUMBER OF NONZEROES ALLOWED IN THE         !ITP02110
!                LINEAR SYSTEM.                                         !ITP02120
!                                                                       !ITP02130
!OUTPUT			                                                        !ITP02140
!                                                                       !ITP02150
!     IA         INTEGER ARRAY OF LENGTH N+1.  SBINI SETS THIS ARRAY    !ITP02160
!                TO -I FOR I = 1 THRU N. IA(N+1) IS SET TO NZ.          !ITP02170
!                                                                       !ITP02180
!     JA         INTEGER ARRAY OF LENGTH NZ. INITIALIZED TO ZERO HERE.  !ITP02190
!                                                                       !ITP02200
!      A         REAL ARRAY OF LENGTH NZ. INITIALIZED TO ZERO HERE.     !ITP02210
!                                                                       !ITP02220
!  IWORK		 INTEGER ARRAY OF LENGTH NZ. INITIALIZED TO ZERO HERE.  !ITP02230
!                                                                       !ITP02240
!***********************************************************************!ITP02250
 INTEGER(K4B), INTENT(IN):: N, NZ										!ITP02260
 INTEGER(K4B), INTENT(OUT):: IA(:), JA(:), IWORK(:)						!ITP02270
 REAL(DP), INTENT(OUT):: A(:)								        	!ITP02280
 INTEGER(K4B):: i                                                       !ITP02290
!***********************************************************************!ITP02300
!                                                                       !ITP02310
    DO i = 1, N															!ITP02320
		IA(i) = -i														!ITP02330
	ENDDO																!ITP02340
    IA(N+1) = NZ														!ITP02350
!	                                                                    !ITP02360
    CALL IVFILL (NZ, JA, 0)												!ITP02370
    CALL IVFILL (NZ, IWORK, 0)											!ITP02380
    CALL  VFILL (NZ, A, 0.D0)											!ITP02390
!                                                                       !ITP02400
!	 					                                                !ITP02410
 END SUBROUTINE SBINI                                                   !ITP02420
!																		!ITP02430
!***********************************************************************!ITP02440
 SUBROUTINE SBSIJ(N,IA,JA,A,IWORK,II,JJ,VALL,MODE,LEVELL,NOUTT,IERR)	!ITP02450
!***********************************************************************!ITP02460
!SUBROUTINE SBSIJ(N,NZ,IA,JA,A,IWORK,II,JJ,VALL,MODE,LEVELL,NOUTT,IERR) !ITP02470
!SBSIJ IS THE SECOND OF A SUITE OF THREE SUBROUTINES TO AID				!ITP02480
!IN THE CONSTRUCTION OF THE IA, JA, A DATA STRUCTURE USED IN			!ITP02490
!ITPACK.																!ITP02500
!                                                                       !ITP02510
!SBSIJ TAKES THE INDIVIDUAL NON ZERO ENTRIES OF THE MATRIX AS			!ITP02520
!GIVEN TO IT AT EACH CALL VIA (II,JJ,VALL) AND INSERTS IT INTO			!ITP02530
!A LINKED LIST REPRESENTATION OF THE SPARSE MATRIX.						!ITP02540
!                                                                       !ITP02550
!EACH ROW OF THE SPARSE MATRIX IS ASSOCIATED WITH A CIRCULAR			!ITP02560
!LINKED LIST BEGINNING AT IA(I). THE LAST ENTERED ELEMENT IN			!ITP02570
!EACH LIST POINTS BACK TO IA(I) WITH THE VALUE -I. THE LINKS			!ITP02580
!ARE STORED IN THE ARRAY IWORK, WHILE JA AND A STORE THE COLUMN			!ITP02590
!NUMBER AND VALUE IN PARALLEL TO IWORK. THE LINKED LISTED ARE			!ITP02600
!STORED BEGINNING AT ENTRY NZ AND WORKING BACKWARDS TOWARDS 1.			!ITP02610
!                                                                       !ITP02620
!PARAMETERS																!ITP02630
!                                                                       !ITP02640
!INPUT																	!ITP02650
!                                                                       !ITP02660
!     N       THE ORDER OF THE LINEAR SYSTEM                            !ITP02670
!                                                                       !ITP02680
!     NZ      THE LENGTH OF THE ARRAYS JA, A, AND IWORK					!ITP02690
!                                                                       !ITP02700
!     II, JJ  THE ROW AND COLUMN NUMBERS OF THE ENTRY OF THE SPARSE     !ITP02710
!             LINEAR SYSTEM TO BE ENTERED IN THE DATA STRUCTURE			!ITP02720
!                                                                       !ITP02730
!     VALL    THE NONZERO VALUE ASSOCIATED WITH (II,JJ)					!ITP02740
!                                                                       !ITP02750
!     MODE    IF THE (I,J) ENTRY HAS ALREADY BEEN SET, MODE SPECIFIES   !ITP02760
!             THE WAY IN WHICH THE ENTRY IS TO BE TREATED.              !ITP02770
!             IF   MODE .LT. 0  LET THE VALUE REMAIN AS IS              !ITP02780
!                       .EQ. 0  RESET IT TO THE NEW VALUE               !ITP02790
!                       .GT. 0  ADD THE NEW VALUE TO THE OLD VALUE      !ITP02800
!                                                                       !ITP02810
!     LEVELL  OUTPUT FILE SWITCH (IF .GE. 0, PRINT OUTPUT FILE)			!ITP02820
!                                                                       !ITP02830
!     NOUTT   OUTPUT FILE NUMBER										!ITP02840
!																		!ITP02850
!INPUT/OUTPUT                                                           !ITP02860
!     IA      INTEGER ARRAY OF LENGTH N+1. THE FIRST N ENTRIES			!ITP02870
!             POINT TO THE BEGINNING OF THE LINKED LIST FOR EACH        !ITP02880
!             ROW. IA(N+1) POINTS TO THE NEXT ENTRY AVAILABLE FOR		!ITP02890
!             STORING THE CURRENT ENTRY INTO THE LINKED LIST.           !ITP02900
!                                                                       !ITP02910
!     JA      INTEGER ARRAY OF LENGTH NZ. JA STORES THE COLUMN			!ITP02920
!             NUMBERS OF THE NONZERO ENTRIES.                           !ITP02930
!                                                                       !ITP02940
!     A       REAL ARRAY OF LENGTH NZ. A STORES THE VALUE OF THE		!ITP02950
!             NONZERO ENTRIES.                                          !ITP02960
!                                                                       !ITP02970
!     IWORK   INTEGER ARRAY OF LENGTH NZ. IWORK STORES THE LINKS.       !ITP02980
!                                                                       !ITP02990
!     IERR    ERROR FLAG. POSSIBLE RETURNS ARE:							!ITP03000
!             IERR =    0  SUCCESSFUL COMPLETION						!ITP03010
!                  =  700  ENTRY WAS ALREADY SET, VALUE HANDLED			!ITP03020
!                          AS SPECIFIED BY MODE.                        !ITP03030
!                  =  701  IMPROPER VALUE OF EITHER I OR J INDEX        !ITP03040
!                  =  702  NO ROOM REMAINING, NZ TOO SMALL.             !ITP03050
!                                                                       !ITP03060
!***********************************************************************!ITP03070
!VARIABLE DECLARATION                                                   !ITP03080
 INTEGER(K4B), INTENT(IN):: N, II, JJ, MODE, LEVELL, NOUTT !, NZ	    !ITP03090
 INTEGER(K4B), INTENT(INOUT):: IA(:), JA(:), IWORK(:)					!ITP03100
 INTEGER(K4B), INTENT(INOUT):: IERR										!ITP03110
 REAL(DP), INTENT(IN):: VALL									        !ITP03120
 REAL(DP), INTENT(INOUT):: A(:)			                                !ITP03130
!																		!ITP03140
 INTEGER(K4B):: LINK, NEXT, NPL1, I, J, LEVEL, NOUT, IER				!ITP03150
 REAL(DP):: VAL, TEMP                                                   !ITP03160
!***********************************************************************!ITP03170
!                                                                       !ITP03180
!CHECK THE VALIDITY OF THE (II,JJ) ENTRY								!ITP03190
!                                                                       !ITP03200
    I = II																!ITP03210
    J = JJ																!ITP03220
    VAL = VALL															!ITP03230
    LEVEL = LEVELL														!ITP03240
    NOUT= NOUTT															!ITP03250
    IER = 0																!ITP03260
    IF ( I .LE. 0  .OR.  I .GT. N ) IER = 701							!ITP03270
    IF ( J .LE. 0  .OR.  J .GT. N ) IER = 701							!ITP03280
    IF (IER .EQ. 0) GOTO 5												!ITP03290
    IF (LEVEL .GE. 0) WRITE(NOUT,3) I,J									!ITP03300
 3	FORMAT(' '												   & 		!ITP03310
         / ' *** FATAL ERROR ********************************' &		!ITP03320
         / ' IMPROPER VALUE FOR PAIR (I,J)=','(',i7,',',i7,')' &		!ITP03330
         / ' ************************************************' &		!ITP03340
         / ' ')															!ITP03350
    GOTO 60																!ITP03360
!                                                                       !ITP03370
!TRAVERSE THE LINK LIST POINTED TO BY IA(I) UNTIL EITHER				!ITP03380
!THE J ENTRY OR THE END OF THE LIST HAS BEEN FOUND.						!ITP03390
!                                                                       !ITP03400
 5  NPL1 = N + 1														!ITP03410
    LINK = IA(I)														!ITP03420
!                                                                       !ITP03430
    IF ( LINK .GT. 0 ) GOTO 10											!ITP03440
																		!ITP03450
!   SPECIAL CASE FOR THE FIRST ENTRY IN THE ROW                         !ITP03460
    NEXT = IA(NPL1)														!ITP03470
    IF ( NEXT .LT. 1 ) GOTO 50	!NZ TOO SMALL							!ITP03480
!                                                                       !ITP03490
    IA(I)       = NEXT													!ITP03500
    JA(NEXT)    = J														!ITP03510
    A(NEXT)     = VAL													!ITP03520
    IWORK(NEXT) = -I													!ITP03530
    IA(NPL1)    = NEXT - 1												!ITP03540
    GOTO 60																!ITP03550
!																		!ITP03560
!	FOLLOW THE LINK LIST UNTIL J OR THE END OF THE LIST IS FOUND		!ITP03570
!                                                                       !ITP03580
 10 IF (JA(LINK) .EQ. J) GOTO 20 !THE POSITION WAS ALREADY STORED		!ITP03590
    IF (IWORK(LINK) .LE. 0) GOTO 30	!END OF LIST FOUND					!ITP03600
    LINK = IWORK(LINK)													!ITP03610
    GOTO 10																!ITP03620
!																		!ITP03630
!ENTRY (I,J) ALREADY HAS BEEN SET. RESET VALUE DEPENDING ON MODE		!ITP03640
!                                                                       !ITP03650
 20 IER = 700															!ITP03660
    IF(MODE .GE. 0) GOTO 22												!ITP03670
!    IF(LEVEL .GE. 1) WRITE(NOUT,21) IER,I,J,A(LINK)					!ITP03680
 21 FORMAT('*** WARNING ************' &									!ITP03690
         / 'IN ITPACK ROUTINE SBSIJ'  &									!ITP03700
         / 'IER = ',I10               &									!ITP03710
         / '(',I10,',',I10,')'        &									!ITP03720
         / 'ENTRY ALREADY SET AND IS LEFT AS ',D15.8 )					!ITP03730
    GOTO 60																!ITP03740
 22 IF(MODE .GE. 1) GOTO 24												!ITP03750
!    IF(LEVEL .GE. 1) WRITE(NOUT,23) IER,I,J,A(LINK),VAL				!ITP03760
 23 FORMAT('*** WARNING ************' &									!ITP03770
         / 'IN ITPACK ROUTINE SBSIJ'  &									!ITP03780
         / 'IER = ',I10				  &									!ITP03790
         / '(',I10,',',I10,')'	  	  &									!ITP03800
         / 'ENTRY ALREADY SET - CURRENT VALUE OF ',D15.8 &				!ITP03810
         / '							RESET TO ',D15.8 )				!ITP03820
    A(LINK) = VAL														!ITP03830
    GOTO 60																!ITP03840
 24 TEMP = A(LINK) + VAL												!ITP03850
!    IF(LEVEL .GE. 1) WRITE(NOUT,25) IER,I,J,A(LINK),TEMP				!ITP03860
 25 FORMAT('*** WARNING ************' &									!ITP03870
         / 'IN ITPACK ROUTINE SBSIJ'  &									!ITP03880
         / 'IER = ',I10               &									!ITP03890
         / '(',I10,',',I10,')'		  &									!ITP03900
         / 'ENTRY ALREADY SET - CURRENT VALUE OF ',D15.8 &				!ITP03910
         / '                            RESET TO ',D15.8 )				!ITP03920
    A(LINK) = TEMP														!ITP03930
    GOTO 60																!ITP03940
!                                                                       !ITP03950
!ENTRY (I,J) HAS NOT BEEN SET. ENTER IT INTO THE LINKED LIST			!ITP03960
!                                                                       !ITP03970
 30 NEXT = IA(NPL1)														!ITP03980
    IF ( NEXT .LT. 1 ) GOTO 50											!ITP03990
!                                                                       !ITP04000
    IWORK(LINK) = NEXT													!ITP04010
    JA(NEXT)    = J														!ITP04020
    A(NEXT)     = VAL													!ITP04030
    IWORK(NEXT) = -I													!ITP04040
    IA(NPL1)    = NEXT - 1												!ITP04050
    GOTO 60																!ITP04060
!                                                                       !ITP04070
!***********************************************************************!ITP04080
!                                                                       !ITP04090
!ERROR TRAP FOR NO ROOM REMAINING										!ITP04100
!                                                                       !ITP04110
 50 IER = 702															!ITP04120
    IF(LEVEL .GE. 0) WRITE(NOUT,55) 									!ITP04130
 55 FORMAT(' ' &														!ITP04140
         / ' *** FATAL ERROR ********************' &					!ITP04150
         / ' NZ TOO SMALL - NO ROOM FOR NEW ENTRY' &					!ITP04160
         / ' ************************************' &					!ITP04170
         / ' ')                                                         !ITP04180
 60 CONTINUE															!ITP04190
 IERR = IER																!ITP04200
!																		!ITP04210
 END SUBROUTINE SBSIJ                                                   !ITP04220
!																		!ITP04230
!***********************************************************************!ITP04240
 SUBROUTINE VFILL(N,V,VAL)                                              !ITP04250
!***********************************************************************!ITP04260
!FILLS A VECTOR, V, WITH A CONSTANT VALUE, VAL.							!ITP04270
!																		!ITP04280
!PARAMETER LIST:                                                        !ITP04290
!				 N      INTEGER LENGTH OF VECTOR V                      !ITP04300
!				 V      REAL VECTOR                                     !ITP04310
!				 VAL    REAL CONSTANT THAT FILLS FIRST N LOCATIONS OF V !ITP04320
!                                                                       !ITP04330
!SPECIFICATIONS FOR DUMMY ARGUMENTS                                     !ITP04340
!                                                                       !ITP04350
 INTEGER(K4B):: N												        !ITP04360
 REAL(DP):: V(N), VAL										        	!ITP04370
!                                                                       !ITP04380
!SPECIFICATIONS FOR LOCAL VARIABLES										!ITP04390
!                                                                       !ITP04400
 INTEGER(K4B):: I, M, MP1												!ITP04410
!                                                                       !ITP04420
    IF (N .LE. 0) RETURN												!ITP04430
!                                                                       !ITP04440
!CLEAN UP LOOP SO REMAINING VECTOR LENGTH IS A MULTIPLE OF 10			!ITP04450
!                                                                       !ITP04460
    M = MOD(N,10)														!ITP04470
    IF (M .EQ. 0) GOTO 20												!ITP04480
	DO 10 I = 1,M														!ITP04490
 10		V(I) = VAL														!ITP04500
    IF (N .LT. 10) RETURN												!ITP04510
!                                                                       !ITP04520
 20 MP1 = M + 1															!ITP04530
    DO 30 I = MP1,N,10													!ITP04540
		V(I  ) = VAL													!ITP04550
        V(I+1) = VAL													!ITP04560
        V(I+2) = VAL													!ITP04570
        V(I+3) = VAL													!ITP04580
        V(I+4) = VAL													!ITP04590
        V(I+5) = VAL													!ITP04600
        V(I+6) = VAL													!ITP04610
        V(I+7) = VAL													!ITP04620
        V(I+8) = VAL													!ITP04630
        V(I+9) = VAL													!ITP04640
 30 CONTINUE															!ITP04650
!                                                                       !ITP04660
!			                                                            !ITP04670
 END SUBROUTINE VFILL                                                   !ITP04680
!																		!ITP04690
!***********************************************************************!ITP04700
 SUBROUTINE IVFILL(N,IV,IVAL)                                           !ITP04710
!***********************************************************************!ITP04720
!																		!ITP04730
!FILLS AN INTEGER VECTOR, IV, WITH AN INTEGER VALUE, IVAL.				!ITP04740
!PARAMETER LIST:                                                        !ITP04750
!			      N INTEGER LENGTH OF VECTOR IV							!ITP04760
!			     IV INTEGER VECTOR										!ITP04770
!			   IVAL INTEGER CONSTANT THAT FILLS FIRST N LOCATIONS OF IV	!ITP04780
!                                                                       !ITP04790
!SPECIFICATIONS FOR DUMMY ARGUMENTS										!ITP04800
!                                                                       !ITP04810
 INTEGER(K4B):: N, IVAL, IV(N)										    !ITP04820
!                                                                       !ITP04830
!SPECIFICATIONS FOR LOCAL VARIABLES										!ITP04840
!                                                                       !ITP04850
 INTEGER(K4B):: I, M, MP1												!ITP04860
!                                                                       !ITP04870
 IF (N .LE. 0) RETURN													!ITP04880
!                                                                       !ITP04890
!CLEAN UP LOOP SO REMAINING VECTOR LENGTH IS A MULTIPLE OF 10			!ITP04900
!                                                                       !ITP04910
	M = MOD(N,10)														!ITP04920
	IF (M .EQ. 0) GOTO 20												!ITP04930
	DO 10 I = 1,M														!ITP04940
 10		IV(I) = IVAL													!ITP04950
	IF (N .LT. 10) RETURN												!ITP04960
!                                                                       !ITP04970
 20	MP1 = M + 1															!ITP04980
    DO 30 I = MP1,N,10													!ITP04990
		IV(I  ) = IVAL													!ITP05000
		IV(I+1) = IVAL													!ITP05010
		IV(I+2) = IVAL													!ITP05020
        IV(I+3) = IVAL													!ITP05030
        IV(I+4) = IVAL													!ITP05040
        IV(I+5) = IVAL													!ITP05050
        IV(I+6) = IVAL													!ITP05060
        IV(I+7) = IVAL													!ITP05070
        IV(I+8) = IVAL													!ITP05080
        IV(I+9) = IVAL													!ITP05090
 30  CONTINUE					                                        !ITP05100
!                                                                       !ITP05110
!							                                            !ITP05120
 END SUBROUTINE IVFILL                                                  !ITP05130
!***********************************************************************

 END MODULE ITPACK
!***