!*****************************************************************************************
!MODULE PARAM---------------------------------------------------------- BY SML. 07/05/2015
!*****************************************************************************************
!
 MODULE PARAM
!
!*****************************************************************************************
!CONTAINS DEFINITION FOR NUMERICAL PRECISION AND OVERALL VARIABLES. THE MODULE MUST BE IN-
!VOKED BY ALL THE RELATED ROUTINES ASSOCIATED TO THE MAIN PROGRAM. IT ALSO CONTAINS A SUB- 
!ROUTINE USED TO PARSE PARAMETER FILE. 
!*****************************************************************************************
!
!*****************
!ATT. MAIN PROGRAM
!*****************
!RDBLK_GIW v3.0
!---
!
 IMPLICIT NONE
!
!------------------------------
!NUMERICAL PRECISION DEFINITION
!------------------------------
 INTEGER, PARAMETER:: K4B=SELECTED_INT_KIND(9)
 INTEGER, PARAMETER:: DP = KIND(0.D0)
!---
!
!----------------
!GLOBAL VARIABLES
!----------------
!PARAMETERS
 INTEGER(K4B):: SEED, NROUND, NZ, N, NRDF, NEFF
! 
!MODEL DEFINITION 
 INTEGER(K4B), ALLOCATABLE:: NEST(:), NLEV(:), TYPO(:), EFFT(:), CORR(:), REPR(:), NLEV1(:)
 LOGICAL, ALLOCATABLE:: TRCK(:,:)
 REAL(DP), ALLOCATABLE:: CVRW(:,:)
!
!ERROR COVARIANCE STRUCTURE
!REAL(DP):: S2ER, S2E0, pPAR, pPAR0
 REAL(DP):: S2EF, pPAR, pPAR0
!
!RANDOM EFFECTS COVARIANCE STRUCTURE
 REAL(DP), ALLOCATABLE::  G(:,:,:), INVG(:,:,:), G0(:,:,:), GTEST(:,:,:)
!
!DEGREES OF BELIEF
 REAL(DP)::  VE1, VEM1, V0, V1
!
!GRIDDY GIBBS SAMPLER
 INTEGER(K4B):: NGRID, MID_VAL
 REAL(DP), ALLOCATABLE:: GRIDD(:)
 REAL(DP):: MIN_VAL, MAX_VAL, INCR, TUN

!
!OUTPUT OPTIONS
 LOGICAL:: PRINTOUT
!---
!
 CONTAINS
!
!*****************************************************************************************
!SUBROUTINE DECODE_PARAM_FILE ----------------------------------------------------- BY SML
!*****************************************************************************************
!
!DECODE THE INSTRUCTIONS GIVEN BY THE USER FOR A DEFINED PROGRAM.
!***
!
 SUBROUTINE DECODE_PARAM_FILE
!
!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE

!LOCAL VARIABLES
 INTEGER(K4B):: i
 
!*********
!PROCEDURE
!*********
!
!OPEN INPUT PARAMETER FILE
 OPEN(37, FILE='param.txt', STATUS="OLD")
!
!SKIP HEADLINES
 DO i = 1, 14
	 READ(37,*)
 ENDDO
!
!----------------- 
!GLOBAL PARAMETERS
!----------------- 
 READ(37,*) SEED
 READ(37,*) NROUND
 READ(37,*) NZ
 READ(37,*) N
 READ(37,*) NRDF
 READ(37,*) NEFF
!---
!
!SKIP HEADLINES
 DO i = 1, 5
	 READ(37,*)
 ENDDO
!---
!
!----------------
!MODEL DEFINITION
!----------------
 ALLOCATE(NEST(NEFF), NLEV(NEFF), TYPO(NEFF), EFFT(NEFF), CORR(NEFF), REPR(NEFF), NLEV1(NEFF))
 READ(37,*) (NEST(i), i = 1, NEFF)
 READ(37,*) (NLEV(i), i = 1, NEFF)
 READ(37,*) (TYPO(i), i = 1, NEFF)
 READ(37,*) (EFFT(i), i = 1, NEFF)
 READ(37,*) (CORR(i), i = 1, NEFF)
 READ(37,*) (REPR(i), i = 1, NEFF)
!
!SKIP HEADLINES
 DO i = 1, 5
	 READ(37,*)
 ENDDO
!---
!
!-----------------
!BAYESIAN ANALISIS
!-----------------
!STARTING VALUES
 ALLOCATE(G(NEFF,2,2), INVG(NEFF,2,2), G0(NEFF,2,2), GTEST(NEFF,2,2))
 READ(37,*) S2EF
 READ(37,*) pPAR0
 G0=0.D0
 GTEST=0.D0
 READ(37,*) G0((NEFF-3),1,1)
 READ(37,*) G0((NEFF-2),1,1)
 READ(37,*) G0((NEFF-1),1,1)
 READ(37,*) G0((NEFF-1),1,2)
 READ(37,*) G0((NEFF-1),2,2)  
 G0((NEFF-1),2,1) = G0((NEFF-1),1,2)
!
!DEGREES OF BELIEF
 READ(37,*)
 READ(37,*)
 READ(37,*) VE1
 READ(37,*) VEM1
 READ(37,*) V0
 READ(37,*) V1
!
!GRIDDY GIBBS SAMPLER
 READ(37,*)
 READ(37,*)
 READ(37,*) NGRID
    IF(MOD(NGRID,2).EQ.0) STOP "PLEASE, SET AN ODD NUMBER OF GRID POINTS. PROGRAM TERMINATED."
 READ(37,*) MIN_VAL, MAX_VAL
 READ(37,*) TUN
!
!SET GLOBAL GRID RANGE PARAMETERS
 MID_VAL = INT((NGRID+1)*0.5D0)
 INCR = (MAX_VAL - MIN_VAL)/(NGRID-1)
 ALLOCATE(GRIDD(NGRID))
!
!SKIP HEADLINES
 DO i = 1, 4
	 READ(37,*)
 ENDDO
!---
!
!--------------
!OUTPUT OPTIONS
!--------------
 READ(37,*) PRINTOUT
!---
!
 CLOSE (UNIT=37)
!
 END SUBROUTINE DECODE_PARAM_FILE
!***
!
 END MODULE PARAM
!***