!----------------
 PROGRAM RDBLK_GS
!----------------
!
!DATE: 07/05/2015
!***
!
!THIS PROGRAM WAS SPECIFICALLY  WRITTEN TO SET-UP AND SOLVE SINGLE-TRAIT MATERNAL ANIMAL MIXED MODEL
!EQUATIONS WHEN DAM-OFFSPRING RESIDUALS ARE CORRELATED AND, THUS, THE ERROR  COVARIANCE MATRIX R HAS
!A BLOCK DIAGONAL STRUCTURE. FURTHER, IT CREATES A MARKOV CHAIN VIA ITERATIVE MONTE CARLO SIMULATION
!USING A SINGLE-SITE SYSTEMATIC GIBBS SAMPLER ALGORITHM. ULTIMATELY, THE MAIN INTEREST IS TO ESTIMA-
!TE THE (CO)VARIANCE COMPONENTS UNDER A BAYESIAN APPROACH.
!
!***VERSION 3.0***
!BASED ON RDBLK v2.0, ALTHOUGH THE GENETIC COVARIANCE  MATRIX IS SAMPLED FROM A GENERALIZED INVERTED
!WISHART (GIW) DISTRIBUTION. THE GIW DISTRIBUTION ALLOWS SETTING DIFFERENT DEGREES OF BELIEF FOR THE
!PRIOR DIRECT BV VARIANCE AND THE PRIOR ADJUSTED MATERNAL BV VARIANCE.
!---
!CODE ADAPTED FROM MAM_GS v1.0, USING PATCHES FROM RDBLK_GS v1.0. IT IS BASED ON AN ALTERNATIVE OPE-
!RATIONAL MODEL WITH A COMMON, KNOWN AND  SMALL WHITE-NOISE ERROR VARIANCE, WHILE INCLUDING AN ADDI- 
!TIONAL ENVIRONMENTAL RANDOM EFFECT TO ACCOUNT FOR THE RESIDUAL DAM - OFFSPRING CORRELATION. THIS A-
!LLOWS SAMPLING WITHOUT SETTING-UP THE MME, THUS ACCELERATING CONSIDERABLY THE PERFORMANCE BY ITERA-
!TION. THE IDEA IS INDEBTED TO DR. ANDRES LEGARRA.
!
!-------------------
!ADDITIONAL FEATURES
!-------------------
! ADMITS BOTH, CATEGORICAL FACTORS & (NESTED) COVARIATES FOR THE FIXED EFFECTS.
! ACCOUNTS FOR INBREEDING THROUGH THE ALGORITHM OF MEUWISSEN AND LUO (1992).
! SUPPORTS POSITIVE SEMI DEFINITE SYSTEMS, ALTHOUGH A REPARAMETERIZATION BLOCK IS AVAILABLE AND RE-
!  COMMENDED TO AVOID NUMERICAL INACCURACIES. ITS ALSO POSSIBLE TO DEFINE AN EFFECT AS A CONTRAST U-
!  SING NEGATIVE VALUES IN THE DATA FILE.
!
!---------------
!TECHNICAL NOTES
!---------------
! SPARSE STORAGE (IJA STRUCTURE) FOR THE MIXED MODEL EQUATIONS.
! USES "CLPS" MODULE TO COLLAPSE A SYSTEM WITH ZERO ROWS AND COLUMNS. 
! USES AN ADAPTIVE, THOUGH NO GROWER, GRIDDY-GIBBS ALGORITHM TO SAMPLE FROM THE UNKNOWN FULL CONDI-
!  TIONAL DENSITY OF THE CORRELATION PARAMETER. 
!
!------------------------------------
!BUGS - ATT. BEWARE SOFTWARE PITFALLS
!------------------------------------
! EXTENSION TO MULTIPLE TRAITS HAS NOT YET BEEN CONSIDERED AND THUS "NTRT" PARAMETER IS SET TO 1.
! RESIDUALS STRONGLY SHRINKED TO ZERO (DEBUGGED 08/06/2010).
! DOUBLE CONTRIBUTION TO RND2 EFFECTS (DEBUGGED 23/09/2011).
!***
!
!AUTHOR: SML * BASED ON THE "COMPUTATIONAL TECHNIQUES IN ANIMAL BREEDING" CLASS NOTES FROM PROFESSOR
!IGNACY MISTZAL, UNIVERSITY OF GEORGIA, ATHENS, USA.
!***
!
!ATT. THIS SYMBOL MEANS "ATTENTION: CHANGE ACCORDING TO THE SPECIFIC MODEL TO BE FITTED".
!***
!
!ATT. COMPILE WITH "PARAM", "ITPACK", "MALG", "SIM", "CLPS" & "SAMPLER" MODULES.
!***
!
!************
!USED MODULES
!************
 USE PARAM
 USE ITPACK
 USE MALG
 USE SIM
 USE CLPS
 USE SAMPLER
!
!**********************************
!SYMBOL DEFINITION (ALPHABETICALLY)
!**********************************
!  ADDJ: SCALAR VARIABLE USED TO ADDRESS CONTRIBUTIONS FOR iid RANDOM EFFECTS. 
! ADDJJ: COEFFICIENT MATRIX ROW NUMBER, USED TO ADDRESS CONTRIBUTIONS FOR RANDOM ANIMAL EFFECTS. 
! ADDKK: COEFFICIENT MATRIX COLUMN NUMBER, USED TO ADDRESS CONTRIBUTIONS FOR RANDOM ANIMAL EFFECTS.
!ADRESS: MATRIX FUNCTION OF (FACTOR,LEVEL,TRAIT) USED TO ADDRESS CONTRIBUTIONS FOR EACH DATA POINT. 
!	ANZ: ACTUAL NUMBER OF NONZEROS.
!  AVEC: REAL, DOUBLE PRECISION VECTOR(NZ). STORES THE VALUE OF THE NONZERO ENTRIES.
!  BETA: VECTOR(NEQ) CONTAINING THE SOLUTIONS OF THE MIXED MODEL EQUATIONS.
!  CLSS: CATEGORICAL FACTOR CODE IDENTIFIER.
!  CORR: VARIABLE USED TO ASCERTAIN CORRELATIONS BETWEEN RANDOM EFFECTS.
!  COVR: COVARIATE CODE IDENTIFIER.
!  CVRW: COVARIATE WEIGHT.
!  DCOL: INTEGER WORK VECTOR(0:NEQ) USED WITHIN COLLAPSE AND GIBBS SUBROUTINES.
!  EFFT: VECTOR WITH THE CATEGORICAL (CLSS) OR COVARIABLE (COVR) CODES OF EACH EFFECT.
! ERROR: INDICATOR VARIABLE USED TO CHECK ALLOCATION OF ALLOCATABLE ARRAYS.
!	ESP: INTEGER WORK VARIABLE USED WITHIN SIM MODULE.
!  FIXX: FIXED EFFECT CODE INDENTIFIER.
!  FLAG: ERROR FLAG WITHIN SIM MODULE.
!     G: RANDOM EFFECTS COVARIANCE MATRICES. TRIDIMENSIONAL ARRAY. FIRST INDEX POINTS TO THE EFFECT.
!    G0: A PRIORI RANDOM EFFECTS COVARIANCE MATRICES ARRAY G.
!  GMAT: SPARSE_MATRIX DERIVED TYPE VARIABLE, USED TO STORE A-INVERSE MATRIX.
!	 IA: VECTOR(NEQ+1). POINTS TO THE FIRST ENTRY OF EACH ROW OF THE COEFFICIENT MATRIX.
!    IP: VECTOR(NEQ) USED TO RETURN THE INVERSE OF THE PERMUTATION RETURNED IN P.
!  IERR: ERROR FLAG WITHIN MODULE ITPACK'S SBSIJ SUBROUTINE.
!INDATA: INCIDENCE VECTOR OF EFFECTS ON A DATA LINE. 
!  INVG: INVERSE OF RANDOM EFFECTS COVARIANCE MATRICES.
!   ISP: VECTOR(NSP) USED FOR WORKING STORAGE WITHIN SIM MODULE.
! IWORK: WORK VECTOR(NZ) USED TO TEMPORARILY STORE LINKED LISTS.
!	 JA: VECTOR(NZ). STORES COLUMN NUMBER OF EACH NON ZERO ELEMENT OF THE COEFFICIENT MATRIX.
!  MAXX: MAXIMUM VALUE POSSIBLE FOR THE DIMENSION OF A RANDOM EFFECT COVARIANCE MATRIX.
!  MISS: MISSING TRAIT CODE IDENTIFIER.
!  MODE: SPECIFIES THE WAY IN WHICH THE ENTRY IS TO BE TREATED, WHEN (I,J) ENTRY HAS ALREADY BEEN SET.
!     N: NUMBER OF ANIMALS IN THE PEDIGREE FILE.
!  NEFF: NUMBER OF FACTORS OR EFFECTS INCLUDING COVARIATES.
!   NEQ: NUMBER OF EQUATIONS.
!  NEST: NESTED COVARIABLE CODE IDENTIFIER.
!  NLEV: NUMBER OF LEVELS OF EACH FACTOR.
! NLEV1: UPDATED NLEV VECTOR AFTER COLLPASING THE SYSTEM.
! NOUTT: OUTPUT FILE UNIT NUMBER WITHIN MODULE ITPACK'S SBSIJ SUBROUTINE.
!  NRDF: NUMBER OF RECORDS IN THE DATA FILE.
!NROUND: NUMBER OF SAMPLER ITERATIONS.
!	NSP: DECLARED DIMENSION OF THE SIM MODULE WORK VECTOR ISP. MUST BE AT LEAST 3*NEQ+4*NZ.
!  NTRT: NUMBER OF TRAITS. WITHIN THIS PROGRAM MUST BE SET TO 1.
!	 NZ: NUMBER OF NON ZEROS IN THE COEFFICIENT MATRIX.
!  NZRW: NON ZERO ROW INDICATOR.
!	  P: PERMUTATION VECTOR(NEQ) USED TO RETURN THE OPTIMAL ORDERING OF MATRIX WITHIN SIM'S MODULE.
! PATH1: PATH SPECIFICATION WITHIN ODRV SUBROUTINE.
! PATH2: PATH SPECIFICATION WITHIN SDRV SUBROUTINE.
!  RANK: ARGUMENT OF GINV SUBROUTINE USED TO CHECK IF MATRICES ARE POSITIVE DEFINITE.
!  RINV: INV-(QMRQm), WHERE Qm IS THE INCIDENCE MATRIX FOR THE mth-MISSING TRAIT PATTERN.
!  RND1: INDENTICALLY AND INDEPENDENTLY DISTRIBUTED (iid) RANDOM EFFECT CODE IDENTIFIER.
!  RND2: RANDOM ANIMAL EFFECT CODE IDENTIFIER.
!  RND3: RANDOM EFFECT WITH SPECIFIC CORRELATION STRUCTURE CODE IDENTIFIER.
!   RSP: VECTOR(NSP) USED FOR WORKING STORAGE WITHIN SIM'S MODULE.
! START: MARKS POSITION OF (FACTOR, LEVEL) COMBINATION FOR EACH DATA POINT ON X'(INV-R)X MATRIX.
!SWITCH: OUTPUT FILE SWITCH WITHIN MODULE ITPACK'S SBSIJ SUBROUTINE. 
!  TITA: VECTOR USED TO STORE THE LOCATION PARAMETERS SAMPLES WITHIN EACH ITERATION.
!  TRCK: LOGICAL MATRIX, USED TO KEEP TRACK OF DEFINED EFFECTS FOR EACH TRAIT.
!  TYPO: VECTOR WITH THE FIXED (FIXX) OR RANDOM (RND1, RND2) CODES OF EACH EFFECT.
!  VALL: SCALAR VARIABLE WITH THE CONTRIBUTION VALUE WITHIN EACH CALL OF MODULE ITPACK. 
!	 WW: PART OF THE COEFFICIENT MATRIX NO ASSOCIATED WITH VARIANCE COMPONENTS.
!    WY: X'Y VECTOR, RHS TIMES ERROR VAIANCE.
!    XY: X'(INV-R)Y VECTOR, RHS.
!     Y: DATA POINT FOR EACH TRAIT.
!  ZROW: BINARY VECTOR(NEQ) USED TO POINT ROWS TO BE DELETED.
!******

!**********************************
!VARIABLE DEFINITION & DECLARATIONS
!**********************************
 IMPLICIT NONE !ALL VARIABLES MUST BE DECLARED.

!------------------
!DERIVED DATA TYPES
!------------------
!"SPARSE_MATRIX_ZERO" IS USED FOR MATRICES WITH ZERO ROWS THAT WILL BE STORED IN FULL IJA-FORMAT.
 TYPE SPARSE_MATRIX_ZERO
	 INTEGER(K4B):: N1, NZ1		!N1: N OF ROWS; NZ1: N OF NONZEROS.
     INTEGER(K4B), ALLOCATABLE::   IA1(:)
     INTEGER(K4B), ALLOCATABLE::   JA1(:)
	 INTEGER(K4B), ALLOCATABLE::  IWK1(:)
	 INTEGER(K4B), ALLOCATABLE:: ZROW1(:)
	 INTEGER(K4B), ALLOCATABLE:: DCOL1(:)
	 REAL(DP):: VALL1				
     REAL(DP), ALLOCATABLE:: A1(:), DUMMY(:)
 END TYPE
!---
!
!"SPARSE_MATRIX" IS USED FOR MATRICES THAT WILL BE STORED IN FULL IJA-FORMAT.
 TYPE SPARSE_MATRIX
	 INTEGER(K4B):: N1, NZ1		!N1: N OF ROWS; NZ1: N OF NONZEROS.
     INTEGER(K4B), ALLOCATABLE::   IA1(:)
     INTEGER(K4B), ALLOCATABLE::   JA1(:)
	 INTEGER(K4B), ALLOCATABLE::  IWK1(:)
	 REAL(DP):: VALL1				
     REAL(DP), ALLOCATABLE:: A1(:)
 END TYPE
!---
!
!"BLOCK" IS USED FOR VARIABLES ASSOCIATED TO A PARTICULAR R_BLOCK.
 TYPE BLOCK
	 INTEGER(K4B):: FK !FK: BLOCK SIZE.
	 REAL(DP), ALLOCATABLE:: S(:,:)
	 REAL(DP), ALLOCATABLE:: R(:,:), RINV(:,:)
 END TYPE
!---
!
!---------------------------
!TYPE DECLARATION STATEMENTS
!---------------------------
 INTEGER(K4B):: NEQ, START
 INTEGER(K4B), PARAMETER:: NTRT=1      
 INTEGER(K4B), PARAMETER:: MAXX=2
 INTEGER(K4B), PARAMETER:: FIXX=0  
 INTEGER(K4B), PARAMETER:: RND1=1
 INTEGER(K4B), PARAMETER:: RND2=2
 INTEGER(K4B), PARAMETER:: RND3=3
 INTEGER(K4B), PARAMETER:: CLSS=0
 INTEGER(K4B), PARAMETER:: COVR=1
 INTEGER(K4B), PARAMETER:: MISS=0
 INTEGER(K4B), PARAMETER:: MODE=5
!---
!
!-----------------------
!FIXED EFFECTS VARIABLES
!-----------------------
 INTEGER(K4B), ALLOCATABLE:: ADRESS(:,:)
 REAL(DP), ALLOCATABLE:: INDATA(:), XY(:), WY(:)
 REAL(DP):: Y(NTRT)
 INTEGER(K4B):: NBKS
 TYPE(BLOCK), ALLOCATABLE:: BK(:)
!---
!
!------------------------------------
!ITPACK, SIM & CLPS MODULES VARIABLES
!------------------------------------
 INTEGER(K4B), ALLOCATABLE:: IA(:), JA(:), IWORK(:), P(:), IP(:), ISP(:), ZROW(:), DCOL(:)
 INTEGER(K4B):: SWITCH, NOUTT, IERR, NSP, PATH1, PATH2, FLAG, ESP
 REAL(DP), ALLOCATABLE:: AVEC(:), BETA(:), RSP(:)
 REAL(DP):: VALL				
!---
!
!-----------------------
!GIBBS SAMPLER VARIABLES
!-----------------------
 INTEGER(K4B):: ERROR, COUNT2
! 
!LOCATION PARAMETERS VARIABLES
 REAL(DP), ALLOCATABLE:: TITA(:)
!
!RANDOM EFFECTS COVARIANCE VARIABLES
 TYPE(SPARSE_MATRIX_ZERO):: WW
 TYPE(SPARSE_MATRIX):: GMAT
!---
!
!-------------------
!OPERATIVE VARIABLES
!-------------------
 INTEGER(K4B):: COUNT, NRDF1, CONT, CUM1, ANZ !COUNTING VARIABLES.
 INTEGER(K4B):: RANK, ADDJ, ADDK, NZRW !WORKING VARIABLES.
 INTEGER(K4B):: i, ii, j, k, l, st, BLK, ROUND !DUMMY VARIABLES.
!---
!
!*********
!PROCEDURE
!*********
 PRINT*, '----------------------'
 PRINT*, 'PROGRAM RDBLK_GIW v3.0'
 PRINT*, '----------------------'
 PRINT*, ''

!DECODE INPUT PARAMETER FILE
 CALL DECODE_PARAM_FILE
 PRINT*, 'PARAMETER FILE SUCCESFULLY READ. START EXECUTION...'
 PRINT*, ''
!---
!
!*********************************************************
!1. INITIALIZATION OF MATRICES, VECTORS & SCALAR VARIABLES
!*********************************************************
 PRINT*, ''
 PRINT*, 'STEP 1. INITIALIZATION.'
 PRINT*, ''
!
!----------------------------------
!SET ITPACK AND SIM MODULES OPTIONS
!----------------------------------
!ATT. SET MODULE ITPACK'S SBSIJ SUBROUTINE OUTPUT FILE OPTIONS.
 SWITCH= 0 !ATT. SET AN INTEGER GREAT OR EQUAL "0" FOR PRINTING THE OUTPUT FILE.
  NOUTT= 6 !ATT. SET THE UNIT NUMBER OF THE OUTPUT FILE. SET '6' TO PRINT ON SCREEN.

!ATT. SET MODULE SIM'S ODRV SUBROUTINE PATH OPTION.
 PATH1= 2 !ATT. SET "2" TO FIND OPTIMAL ORDERING AND REORDER SYMMETRICALLY STORED MATRIX. OTHER
		  !OPTIONS ARE AVAILABLE (SEE LINE "SIM00990" IN SIM'S MODULE).
 PATH2= 1 !ATT. SET "1" TO PERFORM THE THREE STAGES OF SDRV. OTHER OPTIONS ARE AVAILABLE (SEE
		  !LINE "SIM08480" IN SIM'S MODULE). 	
!---
!
!--------------------------------------
!ALLOCATE BOUNDS & INITIALIZE VARIABLES
!--------------------------------------
 NEQ = NTRT*SUM(NLEV)

 ALLOCATE (XY(NEQ), INDATA(NEFF), ADRESS(NEFF, NTRT), IA(NEQ+1), JA(NZ), IWORK(NZ), AVEC(NZ),&
		   ZROW(NEQ), DCOL(NEQ), BETA(NEQ), WY(NEQ), TRCK(NEFF,NTRT), CVRW(NEFF,NTRT))

 XY = 0; WY = 0; INDATA=0; ADRESS = 0; CVRW = 0; BETA = 0; NBKS=0; NRDF1=0; TRCK = .TRUE.
!
 CALL SBINI(NEQ,NZ,IA,JA,AVEC,IWORK)
!
!---------------------------------------------
!INITILIZATION OF DERIVED-DATA TYPE COMPONENTS
!---------------------------------------------
!SET WW MATRIX COMPONENTS AND ALLOCATE SPARSE-STORAGE VARIABLES 
 WW%N1=NEQ
 WW%NZ1=NZ !ATT. SET THE ESTIMATED NUMBER OF NON ZERO ENTRIES FOR WW MATRIX

 ALLOCATE(WW%IA1(WW%N1+1),WW%JA1(WW%NZ1),WW%IWK1(WW%NZ1),WW%ZROW1(WW%N1),WW%DCOL1(WW%N1),&
		  WW%A1(WW%NZ1),WW%DUMMY(WW%N1),STAT=ERROR)
		  IF(ERROR.NE.0) STOP "WW ALLOCATION ERROR" !CHECK FOR ALLOCATION ERROR.

 CALL SBINI(WW%N1,WW%NZ1,WW%IA1,WW%JA1,WW%A1,WW%IWK1)

!SET GMAT MATRIX COMPONENTS AND ALLOCATE SPARSE-STORAGE VARIABLES 
 GMAT%N1 = N
 GMAT%NZ1=INT(NZ*0.2D0) !ATT. SET THE ESTIMATED NUMBER OF NON ZERO ENTRIES FOR GMAT MATRIX

 ALLOCATE(GMAT%IA1(GMAT%N1+1), GMAT%JA1(GMAT%NZ1), GMAT%IWK1(GMAT%NZ1),&
		  GMAT%A1(GMAT%NZ1), STAT=ERROR)
	 	  IF(ERROR.NE.0) STOP "GMAT ALLOCATION ERROR" !CHECK FOR ALLOCATION ERROR.

 CALL SBINI(GMAT%N1,GMAT%NZ1,GMAT%IA1,GMAT%JA1,GMAT%A1,GMAT%IWK1)
!---
!
!-----------------------------------------
!CREATE OPERATIVE INTERNAL INPUT DATA FILE
!-----------------------------------------
 CALL FILE_PREP
!---
!
!--------------------
!OPEN OPERATIVE FILES
!--------------------
 OPEN (21, FILE='DAT1.sal',STATUS='OLD')
 OPEN (25, FILE='CVC_CHAIN.txt',STATUS='REPLACE')  !ATT. NAME OF THE CVC OUTPUT FILE.
!---
!
 PRINT*, ''
 PRINT*, 'NUMBER OF EQUATIONS:', NEQ
 PRINT*, ''
!
!******************************
!2. FIXED EFFECTS CONTRIBUTIONS
!******************************
 PRINT*, '***'
 PRINT*, 'STEP 2. FIXED EFFECTS CONTRIBUTIONS.'
 PRINT*, ''

 COUNT = 1

!---------------------
!CYCLE ALONG DATA FILE
!---------------------
 DO

	!--------------
	!READ DATA LINE
	!--------------
  	 READ(21, *, IOSTAT=st) INDATA, Y

		!-----------
		!CHECK BLOCK
		!-----------
		 IF (st==0) THEN
			 NRDF1 = NRDF1 + 1 !COUNTS NUMBER OF RECORDS
	         IF(NRDF1.eq.(COUNT*1000)) THEN
		         WRITE(*,'(i9,2x,a12)') NRDF1, 'RECORDS READ'
		         COUNT = COUNT + 1
			 ENDIF
		 ELSEIF (st.EQ.-1) THEN !BY DEFAULT, st HAS VALUE -1 IF REACHES DE END OF FILE.
			 PRINT*, ''
			 PRINT*, 'RECORDS IN DATA FILE:', NRDF1
			 EXIT
		 ELSE
			 PRINT*, 'BEWARE! ERROR IN LINE', NRDF1 
			 PRINT*, ''
			 STOP "FATAL ERROR: EXECUTION STOPPED"
		 END IF
		!---

    !----------------------
    !REPARAMERIZATION BLOCK
    !----------------------
    !ATT. SYSTEM SHOULD BE POSITIVE DEFINITE. AVOID DEPENDENCIES SETTING SOME ROWS TO ZERO.
     DO k = 1, NEFF
         IF( REPR(k) .NE. 0) THEN
             IF (INDATA(k)==NLEV(k)) INDATA(k) = 0.D0 !SET TO ZERO THE LAST LEVEL OF THE EFFECT.  
         ENDIF
     ENDDO
	!---
	!
	!--------------
	!ADRESS SETTING
	!--------------
	 CALL SET_ADDRESS 
	!---
	!
   	!--------------------------------
	!COEFFICIENT MATRIX CONTRIBUTIONS
	!--------------------------------
	 DO i = 1, NEFF
		 DO k = 1, NTRT

			!IF THE EFFECT IS NOT DEFINED FOR THE TRAIT:
			 IF (.NOT.TRCK(i,k)) CYCLE

			 DO j = 1, NEFF
				 DO l = 1, NTRT

					!IF THE EFFECT IS NOT DEFINED FOR THE TRAIT:
					 IF (.NOT.TRCK(j,l)) CYCLE

					!----------------------------
					!FULL IJA-STORE FOR WW MATRIX
					!----------------------------
					 IF(ADRESS(i,k).NE.0 .AND. ADRESS(j,l).NE.0) THEN 

				         WW%VALL1 = CVRW(i,k)*CVRW(j,l)*(1/S2EF) 				

						 CALL SBSIJ(WW%N1,WW%IA1,WW%JA1,WW%A1,WW%IWK1,ADRESS(i,k),ADRESS(j,l),&
						            WW%VALL1,MODE,SWITCH,NOUTT,IERR)
									IF(IERR==701 .OR. IERR==702) STOP
					 ENDIF
					!---
					!
					!-------------------------------------------------------
					!UPPER TRIANGULAR IJA-STORAGE FOR MME COEFFICIENT MATRIX
					!-------------------------------------------------------
					 IF(ADRESS(i,k).LE.ADRESS(j,l) .AND. ADRESS(i,k).NE.0 &
					   .AND. ADRESS(j,l).NE.0) THEN

						 VALL = CVRW(i,k)*CVRW(j,l)*(1/S2EF)

						 CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,ADRESS(i,k),ADRESS(j,l),&
									VALL,MODE,SWITCH,NOUTT,IERR)
									IF (IERR==701 .OR. IERR==702) STOP
					 ENDIF
					!---

				 ENDDO
			 ENDDO
			!
			!----------------------
			!SET UP RIGHT HAND SIDE
			!----------------------
			 DO l = 1, NTRT
				 XY(ADRESS(i,k)) = XY(ADRESS(i,k)) + Y(l)*CVRW(i,k)*(1/S2EF)
				 WY(ADRESS(i,k)) = WY(ADRESS(i,k)) + Y(l)*CVRW(i,k)*(1/S2EF)
			 ENDDO
			!---
			 
		 ENDDO
	 ENDDO 
	!---

 ENDDO
!---
!
!*******************************
!3. RANDOM EFFECTS CONTRIBUTIONS
!*******************************
 PRINT*, ''
 PRINT*, '***'
 PRINT*, 'STEP 3. RANDOM EFFECTS CONTRIBUTIONS.'
 PRINT*, ''
!
!-------------------
!CYCLE AMONG EFFECTS
!-------------------
 DO ii = 1, NEFF

	 IF (CORR(ii).EQ.0) CYCLE

	!-----------------------------------------
	!INVERT RANDOM EFFECTS COVARIANCE MATRICES
	!-----------------------------------------
	 INVG = G0
	 CALL GINV(INVG(ii,:,:),CORR(ii)*NTRT,MAXX,1.D-12,RANK)
	 
	!SECURITY DEVICE: CHECK OUT IF G(ii,:,:) IS POSITIVE DEFINITE.
	 IF(RANK .LT. (CORR(ii)*NTRT)) THEN
		 PRINT*, ''
		 PRINT*, '-----------'
		 WRITE(*,'(a24,i2,a27)') ' FATAL ERROR: MATRIX G0(', ii, ') IS NOT POSITIVE DEFINITE.'
		 PRINT*, '-----------'
		 PRINT*, ''
		  STOP
	 ENDIF
	!---
	!
	!-----------------------------
	!SELECT THE RANDOM EFFECT TYPE
	!-----------------------------
	 SELECT CASE (TYPO(ii))
	
		!------------
 		!FIXED EFFECT
		!------------
		 CASE (FIXX)

			 CONTINUE !SKIP EFFECT

		!-----------------
		!RANDOM IID EFFECT
		!-----------------
		 CASE (RND1)

			!------------------
			!CYCLE AMONG TRAITS
			!------------------
			 DO i = 1, NTRT

				!IF THE EFFECT IS NOT DEFINED FOR THE TRAIT:
				 IF (.NOT.TRCK(ii,i)) CYCLE

				!-----------------------------------------
				!CYCLE AMONG LEVEL WITHIN TRAIT AND EFFECT
				!-----------------------------------------
				 START = SUM(NLEV(1:ii-1))*NTRT + i

				 DO j = 1, NLEV(ii)

					!ADRESS SETTING.
					 ADDJ =  START + (j - 1)*NTRT

					!ENTRIES CALCULATION.
				     CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,ADDJ,ADDJ,INVG(ii,i,i),MODE,SWITCH,NOUTT,IERR)
								IF(IERR==701 .OR. IERR==702) STOP 

				 ENDDO
				!---

			 ENDDO
			!---

		!--------------------
		!RANDOM ANIMAL EFFECT
		!--------------------
		 CASE (RND2)

			 CALL INVA(ii)

		!------------------------------------------------
		!RANDOM EFFECT WITH SPECIFIC COVARIANCE STRUCTURE
		!------------------------------------------------
		 CASE (RND3)

			 START = SUM(NLEV(1:ii-1)) + 1

			 DO BLK = 1, NBKS

				!----------------------------------------
				!SET-UP R-MATRICES WITH STORED S-MATRICES
				!----------------------------------------
				 DO i = 1, BK(BLK)%FK
					 BK(BLK)%R(i,i) = 1.D0 
				 ENDDO
				 BK(BLK)%R = BK(BLK)%R + BK(BLK)%S*pPAR0
				!---
				!
		   		!---------------
				!INVERT R-MATRIX
				!---------------
				 IF(BLK.NE.NBKS) THEN

					 BK(BLK)%RINV = BK(BLK)%R
	 
					 CALL GINV(BK(BLK)%RINV,BK(BLK)%FK,BK(BLK)%FK,1.D-12,RANK)

					!SECURITY DEVICE: CHECK OUT IF SCALE MATRIX IS POSITIVE DEFINITE.
					 IF(RANK .LT. BK(BLK)%FK) THEN
						 PRINT*, ''
						 PRINT*, '-----------'
						 WRITE(*,'(a30,i3,a27)') 'FATAL ERROR: RINV FOR FAMILY ', BLK,&
												 ' IS NOT POSITIVE DEFINITE.' 
						 PRINT*, '-----------'
						 PRINT*, ''
						  STOP
					 ENDIF

				 ELSE
					 DO i = 1, BK(BLK)%FK	
						 BK(BLK)%RINV(i,i) = 1.D0
					 ENDDO
				 ENDIF	
				!---
				!
		   		!-------------
				!CONTRIBUTIONS
				!-------------
   				 IF(BLK.NE.NBKS) THEN

					 DO j = 1, BK(BLK)%FK
						 DO k = 1, BK(BLK)%FK

							!ADRESS SETTING.
							 ADDJ =  START + (j - 1)
							 ADDK =  START + (k - 1)

							!ONLY COMPUTE UPPER DIAGONAL CONTRIBUTIONS
							 IF(ADDJ.LE.ADDK) THEN

								!ENTRIES CALCULATION.
								 VALL = BK(BLK)%RINV(j,k)*INVG(ii,1,1)
							     CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,ADDJ,ADDK,VALL,MODE,&
											SWITCH,NOUTT,IERR)
											IF(IERR==701 .OR. IERR==702) STOP 
							 ENDIF

						 ENDDO
					 ENDDO

				 ELSE

					 DO j = 1, BK(BLK)%FK

						!ADRESS SETTING.
						 ADDJ =  START + (j - 1)

						!ENTRIES CALCULATION.
						 VALL = BK(BLK)%RINV(j,j)*INVG(ii,1,1)
					     CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,ADDJ,ADDJ,VALL,MODE,&
									SWITCH,NOUTT,IERR)
									IF(IERR==701 .OR. IERR==702) STOP 

					 ENDDO

				 ENDIF	
				 START = START + BK(BLK)%FK

			 ENDDO
			!---

		!----
		!ELSE
		!----
		 CASE DEFAULT
			 PRINT*, 'UNIMPLEMENTED RANDOM EFFECT', TYPO(ii)
			 PRINT*, ''
			 STOP "FATAL ERROR: EXECUTION STOPPED"
			!---

	 END SELECT
	!---

 ENDDO
!---
!
!*****************************
!4. BUILD FINAL DATA STRUCTURE
!*****************************
 PRINT*, ''
 PRINT*, '***'
 PRINT*, 'STEP 4. BUILDING FINAL DATA STRUCTURE.'
 PRINT*, ''

!WHOLE ARRAY ASSIGNMENT FOR UPDATED NLEV1 VECTOR.
 NLEV1 = NLEV 

!-----------------------
!SEARCHING FOR ZERO ROWS
!-----------------------
 ZROW = 0
 DCOL = 0
 CUM1 = 0
 NZRW = 0
 CONT = 0

!SEARCH AND TAG ZERO ROWS
 DO i = 1, NEFF
	 DO j = 1, NLEV(i)
		 CONT = CONT + 1
		 IF (IA(CONT).LE.0) THEN
			 NZRW = 1
			 CUM1 = CUM1 + 1 
			 ZROW(CONT) = 1
			 NLEV1(i) = NLEV1(i)-1 !UPDATING NLEV VECTOR.
		 ENDIF
		 DCOL(CONT) = CUM1
	 ENDDO	
 ENDDO
!---
!
!-----------------------------------------
!BUILD SPARSE-STORAGE FINAL DATA STRUCTURE
!-----------------------------------------
 CALL SBEND(NEQ,NZ,IA,JA,AVEC,IWORK)
 CALL SBEND(WW%N1,WW%NZ1,WW%IA1,WW%JA1,WW%A1,WW%IWK1)
 CALL SBEND(GMAT%N1,GMAT%NZ1,GMAT%IA1,GMAT%JA1,GMAT%A1,GMAT%IWK1) 
!---
!
!---------------------------
!COLLAPSE COEFFICIENT MATRIX
!---------------------------
 IF (NZRW.NE.0) THEN

	 PRINT*, 'ZERO ROWS FOUND. COLLAPSING THE SYSTEM...'

		!MME FINAL STRUCTURE
		 CALL COLLAPSE(NEQ, IA, JA, XY, ZROW, DCOL(1:NEQ))

		!WW FINAL STRUCTURE
	 	 WW%ZROW1 = ZROW
		 WW%DCOL1 = DCOL

		 CALL COLLAPSE(WW%N1,WW%IA1, WW%JA1, WY, WW%ZROW1, WW%DCOL1)
		!---

	 PRINT*, 'REMAINING EQUATIONS:', NEQ
				   
 ENDIF
 IF(NEQ.NE.SUM(NLEV1)) STOP "CHECK NLEV UPDATING" 
!---
! 
!-----------------------------------------
!SET THE ACTUAL NUMBER OF NON ZERO ENTRIES
!-----------------------------------------
 PRINT*, ''
 PRINT*, 'SETTING THE ACTUAL NUMBER OF NONZERO ENTRIES...'

 ANZ = IA(NEQ+1) - 1
 PRINT*, 'No. NONZERO ENTRIES:', ANZ
!---
!
!************************************
!5. SOLUTION OF MIXED MODEL EQUATIONS
!************************************
 PRINT*, ''
 PRINT*, '***'
 PRINT*, 'STEP 5. SOLVING MIXED MODEL EQUATIONS.'
 PRINT*, ''
!
!-----------------------------------------------
!INITIALLIZE AND ALLOCATE SOLVER'S ARRAYS BOUNDS
!-----------------------------------------------
 NSP = 3*NEQ + 6*ANZ
 ALLOCATE (P(NEQ), IP(NEQ), ISP(NSP), RSP(NSP))
!---
!
!----------------
!SOLVE THE SYSTEM
!----------------
!SEARCH FOR OPTIMAL ORDERING.
 CALL ODRV(NEQ,IA(1:NEQ+1),JA(1:ANZ),AVEC(1:ANZ),P,IP,NSP,ISP,PATH1,FLAG)

	!AN ERROR HAS OCCURRED. STOP PROCESS.
	 IF (FLAG.NE.0) THEN
		 PRINT*, 'ERROR IN ODRV'
		 STOP
	 ENDIF
	!---

!SOLVE EQUATION'S SYSTEM.
 CALL SDRV(NEQ,P,IP,IA(1:NEQ+1),JA(1:ANZ),AVEC(1:ANZ),XY(1:NEQ),BETA(1:NEQ),&
		   NSP,ISP,RSP,ESP,PATH2,FLAG)

	!AN ERROR HAS OCCURRED. STOP PROCESS.
	 IF (FLAG.NE.0) THEN
		 PRINT*, 'ERROR IN SDRV'
		 PRINT*, 'FLAG:', FLAG
	     STOP
	 ENDIF
	!---

!*****************
!6. GIBBS SAMPLING
!*****************
 PRINT*, '***'
 PRINT*, 'STEP 6. SAMPLING.'
 PRINT*, ''
!
!----------------------------------
!ALLOCATE LOCATION PARAMETER VECTOR 
!----------------------------------
 ALLOCATE(TITA(NEQ)); TITA = BETA
!---
!
!------------------------------------------------------------
!INITIALIZE COVARIANCE STRUCTURE PARAMETERS BEFORE ITERATIONS
!------------------------------------------------------------
 G = G0
 pPAR = pPAR0
!---
!
!--------------------
!GIBBS SAMPLER CYCLES
!--------------------
 COUNT2 = 1
  
 DO ROUND = 1, NROUND

	!------------------------------
	!PRINT THE NUMBER OF ITERATIONS 
	!------------------------------
	!ATT. SET THE ROUND COUNTER
     IF(ROUND.eq.(COUNT2*1)) THEN
         WRITE(*,'(i9,2x,a12)') ROUND, 'TH ITERATION'
         COUNT2 = COUNT2 + 1
	 ENDIF
	!---
	!
	!--------------------------------
	!COMPUTE UPDATED INVERSE G-MATRIX
	!--------------------------------
	 INVG = G !SET INVG WITH UPDATED G MATRIX

	 DO ii = 1, NEFF

		 IF (CORR(ii).EQ.0) CYCLE

		!INVERT UPDATED RANDOM EFFECTS COVARIANCE MATRICES.
		 CALL GINV(INVG(ii,:,:),CORR(ii)*NTRT,MAXX,1.D-12,RANK)

		!SECURITY DEVICE: CHECK OUT IF G(ii,:,:) IS POSITIVE DEFINITE.
		 IF(RANK .LT.(CORR(ii)*NTRT)) THEN
			 PRINT*, ''
			 PRINT*, '-----------'
			 WRITE(*,'(a23,i1,a27)') 'FATAL ERROR: MATRIX G(', ii, ') IS NOT POSITIVE DEFINITE.' 
			 PRINT*, '-----------'
			 PRINT*, ''
			  STOP
		 ENDIF

	 ENDDO
	!---
	!
	!-----------------------------
	!CALL GIBBS SAMPLER SUBROUTINE
	!-----------------------------
	 CALL GIBBS(ROUND)
	!---

 ENDDO
!---
!
!---------------------
!CLOSE OPERATIVE FILES
!---------------------
 CLOSE(UNIT=21)
 CLOSE(UNIT=25)
!---

!ATT. COMMENT THIS STATEMENT TO PRINT OUT SOLUTIONS BEFORE FIRST ITERATION.
 IF(.NOT. PRINTOUT) STOP "PROGRAM TERMINATED - NO SOLUTIONS."
!***

!*********
!7. OUTPUT
!*********
 PRINT*, ''
 PRINT*, '***'
 PRINT*, 'STEP 7. WRITTING OUTPUT FILE.'

!--------------------------------------
!RETRIEVE SOLUTION'S ORIGINAL STRUCTURE
!--------------------------------------
 CALL RETRIEVE(NEQ, ZROW, BETA, 0.D0)

!-------------------------------------
!OPEN OUTPUT FILES AND PRINT HEADLINES
!-------------------------------------
 OPEN (11,file='RDBLK TEST - SOLS.txt',STATUS='REPLACE') !ATT. NAME OF OUTPUT FILE. 

 WRITE(11,'(a14)') '++++++++++++++'
 WRITE(11,'(a14)') 'SOLUTIONS FILE'
 WRITE(11,'(a14)') '++++++++++++++'
 WRITE(11,*) ''

 WRITE(11,'(a6,2x,a5,2x,a5,2x,a15)') 'EFFECT', 'LEVEL', 'TRAIT', '      SOLUTIONS'
 WRITE(11,'(a6,2x,a5,2x,a5,2x,a15)') '------', '-----', '-----', '      ---------'

 START = 0

!-------------------
!CYCLE AMONG EFFECTS
!-------------------
 DO i = 1, NEFF

	!--------------
	!PRINT HEADINGS
	!--------------
	 IF (TYPO(i)==RND1 .OR. TYPO(i)==RND3) THEN
		 WRITE(11,*) ''
		 WRITE(11,'(a6,2x,a5,2x,a5,2x,a15)') 'EFFECT', 'LEVEL', 'TRAIT', '   PREDICTIONS'
		 WRITE(11,'(a6,2x,a5,2x,a5,2x,a15)') '------', '-----', '-----', '   -----------'
	 ELSEIF (TYPO(i)==RND2) THEN
		 WRITE(11,*) ''
		 WRITE(11,'(a6,2x,a5,2x,a5,2x,a15)') 'EFFECT', '   ID', 'TRAIT', 'BREEDING VALUE'
		 WRITE(11,'(a6,2x,a5,2x,a5,2x,a15)') '------', '-----', '-----', '--------------'
	 ENDIF

	!-------------
	!PRINT RESULTS
	!-------------
	 DO j = 1, NLEV(i)
		 DO k = 1, NTRT
			 START = SUM(NLEV(1:i-1))*NTRT + (j-1) + k
			 WRITE(11,'(i6,2x,i5,2x,i5,2x,f15.4)') i, j, k, BETA(START)
		 ENDDO
	 ENDDO
	 WRITE(11,'(a37)') '-------------------------------------'
	 WRITE(11,*) ''

 ENDDO
 CLOSE (11)
!---

 PRINT*, ''
 PRINT*, '***'
 STOP "PROGRAM TERMINATED."

 CONTAINS
!
!*****************************************************************************************
!INTERNAL SUBROUTINE FILE_PREP ---------------------------------------- BY SML. 28/05/2010
!*****************************************************************************************
!
!REORDERS INPUT DATA FILE RECORDS IN A WITHIN-BLOCK FASHION, AND COMPUTES BOTH, THE NUMBER
!OF BLOCKS AND THE SIZE OF EACH ONE OF THEM. FINALLY, SET-UP AND STORES THE COVARIANCE MA-
!TRICES STRUCTURE, Sk, AND PRINTS-OUT THE OPERATIVE INTERNAL DATA FILE.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!READS AND STORES INPUT DATA FILE. THIS FILE MUST BE ORDERED SUCH THAT PARENTS ALWAYS PRE-
!CEDE THEIR OFFSPRING. ASSIGN TO EACH INDIVIDUAL A NUMERICAL CONSECUTIVE CODE ACCORDING TO
!THE BLOCK IT BELONGS. SET-UP AND STORES S-MATRICES ACCORDING TO DAM-OFFSPRING PAIRS.
!*****************************************************************************************
!
!ATT. BEWARE! THIS SUBROUTINE MUST BE COMPILED WITHIN ANY "RDBLK_GS" PROGRAM.
!
 SUBROUTINE FILE_PREP

!LOCAL VARIABLES 
 INTEGER(K4B):: DAMI !BINARY VARIABLE USED TO SET THE STATUS OF A FEMALE.
 INTEGER(K4B):: WKI(NRDF)	!FOR EACH INDIVIDUAL, STORE THE BLOCK-ID IT BELONGS.
 
 REAL(DP), ALLOCATABLE:: DATALINE(:,:)

 INTEGER(K4B):: COLM, COUNT, NRBK, ERR !COUNTING VARIABLES.
 INTEGER(K4B):: i, j, BLK, st1 !DUMMY VARIABLES.

!*********
!PROCEDURE
!*********

!--------------------------------
!STEP 1.READ AND STORE DATA BLOCK
!--------------------------------
 COLM = NEFF

 ALLOCATE(DATALINE(NRDF,COLM))

 WKI = 0; NRBK = 1

 OPEN (1,FILE='dat.txt',STATUS='OLD') !ATT. NAME OF THE INPUT DATA FILE. 

 DO i = 1, NRDF

	!-----------------
	!DATA LINE READING
	!-----------------
	!ATT. "SEX" IT IS ASSUMED TO BE IN THE FIRST FIELD WITHIN THE DATA FILE.
	!ATT. "IDM" AND "ID" ARE ASSUMED TO BE THE "NEFF-1" AND "NEFF-2" POSITIONS IN THE DATA LINE.
  	 READ(1, *, IOSTAT=st1) DATALINE(i,:)

		!DATA READING CHECK BLOCK
		 IF (st1==0) THEN
			 CONTINUE
		 ELSEIF (st1.EQ.-1) THEN !BY DEFAULT, st1 HAS VALUE -1 IF IT REACHES THE END OF FILE.
			 PRINT*, ''
			 PRINT*, 'BEWARE! UNEXPECTED END OF FILE. CHECK NRDF PARAMETER' 
			 STOP
		 ELSE
			 PRINT*, 'BEWARE! ERROR', st1,' IN LINE', i 
			 PRINT*, ''
			 STOP "FATAL ERROR: EXECUTION STOPPED"
		 END IF
		!---
 ENDDO
!---
!
!--------------------------
!STEP 2. FILL-UP WKI VECTOR
!--------------------------
 DO i = 1, NRDF

	 DAMI = 0

	!IF THE INDIVIDUAL IS A MALE, CYCLE.
	 IF (DATALINE(i,1).NE.2.D0) CYCLE

	!SET WKI FOR THE FIRST DAM WITHIN THE BLOCK:
	 IF (WKI(i).EQ.0) THEN
		 WKI(i) = NRBK
		 NRBK = NRBK + 1
	 ELSE
		 DAMI = 1
	 ENDIF

	!SEARCH FOR OFFSPRING, AND SET THE CORRESPONDING WKI
	 DO j = 1, NRDF
		 IF (DATALINE(j,COLM-1).EQ. DATALINE(i,COLM-2)) THEN
			 WKI(j) = WKI(i)
			 DAMI = 1
		 ELSE
			 CYCLE		 
		 ENDIF
	 ENDDO

	!CORRECT IF THE FEMALE IS NEITHER A DAM, NEITHER A DAM'S OFFSPRING
	 IF(DAMI.EQ.0) THEN
		 WKI(i)=0
		 NRBK = NRBK - 1
	 ENDIF

 ENDDO

 NBKS =	NRBK
  
!SEND UNLINKED INDIVIDUALS TO THE LAST BLOCK
 DO i = 1, NRDF
	 IF(WKI(i).EQ.0) WKI(i)=NBKS
 ENDDO
!
!PRINT ON SCREEN THE NUMEBR OF MATERNAL FAMILIES
 PRINT*, ''
 PRINT*, '---------------------'
 WRITE(*,'(a23,1x,i4)') ' No. MATERNAL FAMILIES:', NRBK
 PRINT*, '---------------------'
 PRINT*, ''
!---
!
!--------------------------------
!STEP 3. PRINT-OUT OPERATIVE FILE
!--------------------------------
!ATT. RESIDUAL EFFECT WILL BE PLACED IN THE "NEFF-2" `POSITION.
 OPEN (21,FILE='DAT1.sal',STATUS='REPLACE')

 ALLOCATE(BK(NBKS))
 
 COUNT = 0
 
 DO j = 1, NBKS
 	 BK(j)%FK=0
	 DO i = 1, NRDF
		 IF(WKI(i).EQ.j) THEN
			 COUNT  = COUNT+1
			 BK(j)%FK = BK(j)%FK + 1
			 WRITE(21,*) DATALINE(i,:COLM-3), 1.D0*COUNT, DATALINE(i,COLM-2:)

		 ELSE
			 CYCLE		 
		 ENDIF 		
	 ENDDO
 ENDDO
 IF(COUNT .NE. NRDF) STOP "FATAL ERROR IN SUBROUTINE FILE_PREP"
 DEALLOCATE(DATALINE)
!---
!
!-------------------------
!STEP 4. SET-UP S-MATRICES
!-------------------------
 REWIND (UNIT=21)
 DO BLK = 1, NBKS

	!-------------------------------------
	!ALLOCATE AND INITILIZE BLOCK MATRICES
	!-------------------------------------
	 ALLOCATE(BK(BLK)%S(BK(BLK)%FK,BK(BLK)%FK),BK(BLK)%R(BK(BLK)%FK,BK(BLK)%FK),&
			  BK(BLK)%RINV(BK(BLK)%FK,BK(BLK)%FK), STAT=ERR)
		 	  IF(ERR.NE.0) STOP "BLOCK ALLOCATION ERROR" !CHECK FOR ALLOCATION ERROR.

	 BK(BLK)%S=0; BK(BLK)%R=0; BK(BLK)%RINV=0 

	 IF(BLK==NBKS) EXIT 
	!---
	!
	!---------------
	!READ BLOCK DATA
	!---------------
	 ALLOCATE(DATALINE(BK(BLK)%FK,COLM+1))

	 DO i = 1, BK(BLK)%FK 
	  	 READ(21, *) DATALINE(i,:)
	 ENDDO
	!---
	!
	!------------------
	!COMPUTE S-MATRICES
	!------------------
	 DO i = 1, BK(BLK)%FK

		 DO j = i+1, BK(BLK)%FK

			!IF j IS DAM OF i:
			 IF (DATALINE(j,COLM).EQ. DATALINE(i,COLM-1)) THEN
				 BK(BLK)%S(i,j) = 1.D0			 !ADDS 1 TO THE CORRESPONDING ENTRY
				 BK(BLK)%S(j,i) = BK(BLK)%S(i,j) !TRANSPOSE RESULT
			 ENDIF

		 ENDDO
		
	 ENDDO
	 DEALLOCATE(DATALINE)
	!---

 ENDDO
 CLOSE (UNIT=1); CLOSE (UNIT=21)
!---

 END SUBROUTINE FILE_PREP  
!***
!
!*****************************************************************************************
!INTERNAL SUBROUTINE INVA---------------------------------------------- BY SML. 17/12/2008
!*****************************************************************************************
!
!RETURNS THE INVERSE ADDITIVE RELATIONSHIP MATRIX, INV-A, CONTRIBUTIONS TO THE MIXED MODEL
!EQUATIONS.
!***
!
!***VERSION 2.2.1***
!SAVE INVA v2.1 BUG. INBRED POPULATIONS. SPARSE STORAGE. DOUBLE PRECISION REAL VARIABLES.
!
!ALGORITHM DESCRIPTION********************************************************************
!BASED ON HENDERSON'S RULES [BIOMETRICS 1976, 32:69] AND MEUWISSEN & LUO'S WORK [GSE 1992,
!24:30]. CODE BASED ON SUBROUTINE "INBRED v2.0".
!*****************************************************************************************
!
!BEWARE! THIS SUBROUTINE MUST BE COMPILED WITHIN "MAM_GS" PROGRAM. 
!***
!
 SUBROUTINE INVA (ii)
!
!INTENT(IN): ii - ANIMAL EFFECT INDEX.
!***
!
!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE
!
!--------------
!DUMMY ARGUMENT
!--------------
 INTEGER(K4B), INTENT(IN):: ii
!---
!
!---------------
!LOCAL VARIABLES 
!---------------
 INTEGER(K4B):: IN, IS, ID, KS, KD !WORK VARIABLES.

	!MEUWISSEN AND LUO'S ALGORITHM VARIABLES
	 INTEGER(K4B):: POINT(1:N)      !POINTS TO THE NEXT ANCESTOR IN THE LINKED LIST.
	 INTEGER(K4B):: PED(1:3, 1:N)   !PEDIGREE
	 REAL(DP):: F(0:N) !INBREEDING COEFFICIENTS.
	 REAL(DP):: D(1:N) !MENDELIAN RESIDUAL VARIANCE CONTRIBUTION TO A-INV.
	 REAL(DP):: L(1:N) !FOR INDIVIDUAL i, ENTRIES (i,j) OF L MATRIX. 
	 REAL(DP):: R      !WORK VARIABLE. ACCUMULATE [0.5*L(i,k)] FOR k= {ANC}.
	 REAL(DP):: FI     !WORK VARIABLE. ACCUMULATE [L(i,j)^2*D(j,j)] FOR j= 1 TO i.

	!CONTRIBUTIONS VARIABLES
	 INTEGER(K4B):: ADDJJ, ADDKK, COUNT1, PFLAG
	 INTEGER(K4B):: P(3)
	 REAL(DP):: WORK(3) = (/1.D0,-0.5D0,-0.5D0/)

 INTEGER(K4B):: i, j, k,  c1, c2, jj, kk !DUMMY VARIABLES.
!---
!
!*********
!PROCEDURE
!*********
 POINT = 0		
  F(0) = -1.D0
	 L =  0.D0

!----------------------
!OPENING PEDIGREE FILES
!----------------------
 OPEN (2,file='ped.txt', STATUS='OLD')   !ATT. NAME OF PEDIGREE FILE. 
 OPEN (3,file='ERROR FILE.txt', STATUS='REPLACE')

 WRITE(3,'(a11)') '+++++++++++'
 WRITE(3,'(a11)') 'ERRORS FILE'
 WRITE(3,'(a11)') '+++++++++++'
 WRITE(3,*) ''
 WRITE(3,'(a24)') 'PEDIGREE INCONSISTENCIES'
 WRITE(3,'(a24)') '------------------------'
 WRITE(3,*) ''
 WRITE(3,'(a7,2x,a12)') '     ID', 'OBSERVATIONS'
 WRITE(3,'(a7,2x,a12)') '-------', '------------'
!---
!
!---------------------
!READING PEDIGREE FILE
!---------------------
 COUNT1 = 1
  PFLAG = 0

!-----------------------
!CYCLE AMONG INDIVIDUALS
!-----------------------
 DO i = 1, N

	!-------------------------------
	!PEDIGREE LINE READING & STORING
	!-------------------------------
	 READ(2, *) PED(:,i)

		!----------------
 		!PEDIGREE COUNTER
		!----------------
         IF(i.eq.(COUNT1*1000)) THEN
			 WRITE(*,'(i9,2x,a18)') i, 'TH INDIVIDUAL READ'
			 COUNT1 = COUNT1 + 1
		 ENDIF
		!---

		!------------------------------
		!CHECK PEDIGREE INCONSISTENCIES
		!------------------------------
		 IF (PED(2,i).GE.PED(1,i)) THEN
			 IF (PED(2,i).EQ.PED(1,i)) THEN
				 WRITE(3,'(i7,2x,a36)') PED(1,i), 'INDIVIDUAL AND ITS SIRE ID ARE EQUAL'
			 ELSE
			 	 WRITE(3,'(i7,2x,a36)') PED(1,i), 'INDIVIDUAL ID IS LOWER THAN ITS SIRE'
			 ENDIF
			 PED(2,i) = 0
			 PFLAG = 1
		 ELSEIF (PED(3,i).GE.PED(1,i)) THEN
			 IF (PED(3,i).EQ.PED(1,i)) THEN
				 WRITE(3,'(i7,2x,a35)') PED(1,i), 'INDIVIDUAL AND ITS DAM ID ARE EQUAL'
			 ELSE
			 	 WRITE(3,'(i7,2x,a35)') PED(1,i), 'INDIVIDUAL ID IS LOWER THAN ITS DAM'
			 ENDIF
			 PED(3,i) = 0
			 PFLAG = 1
		 ELSE
			 CONTINUE
		 ENDIF
		!---

	!----------------------------------
	!MEUWISSEN & LUO'S ALGORITHM KERNEL
	!----------------------------------
	 IN = PED(1,i)
	 IS = PED(2,i)
	 ID = PED(3,i)
	
	 PED(2,i) = MAX(IS,ID)
	 PED(3,i) = MIN(IS,ID)
	 
     D(i) = 0.5D0 - 0.25D0*(F(IS)+F(ID))

	!IF EITHER OF THE PARENTS IN UNKNOWN:
	 IF (IS.EQ.0 .OR. ID.EQ.0) THEN

	 	 F(i) = 0.D0 !INDIVIDUAL IS NOT INBRED

	!IF INDIVIDUALS i AND (i-1) ARE FULL SIBS:
	 ELSEIF (PED(2, i-1).EQ.PED(2,i) .AND. PED(3, i-1).EQ.PED(3,i)) THEN

		 F(i) = F(i-1) !BOTH HAS THE SAME INBREEDING COEFFICIENT

	 ELSE

		   FI = -1.D0
		 L(i) =  1.D0
		    j =  PED(1,i)

		 DO WHILE (j.NE.0)
		 
			  k = j
 			  R = 0.5D0*L(k)
			 KS = PED(2,k)
			 KD = PED(3,k)

			!IF ANCESTOR'S SIRE IS KNOWN:
			 IF(KS.GT.0) THEN

				 DO WHILE (POINT(k).GT.KS)
				 k = POINT(k)
				 ENDDO

				 L(KS) = L(KS) + R

				!IF ANCESTOR'S SIRE IS NOT NEXT IN THE ANCESTOR'S LIST: 
				 IF (KS.NE.POINT(k)) THEN
 				 POINT(KS) = POINT(k)
				 POINT(k)  = KS
				 ENDIF

				!IF ANCESTOR'S DAM IS KNOWN:
				 IF(KD.GT.0) THEN

					 DO WHILE (POINT(K).GT.KD)
					 k = POINT(k)
					 ENDDO

					 L(KD) = L(KD) + R

					!IF ANCESTOR'S DAM IS NOT NEXT IN THE ANCESTOR'S LIST:
					 IF (KD.NE.POINT(K)) THEN
 					 POINT(KD) = POINT(k)
					 POINT(k)  = KD
					 ENDIF

				 ENDIF
				!***

			 ENDIF
			!***

			 FI = FI + L(j)*L(j)*D(j)
			 L(j) = 0.D0
			 
			 k = j
			 j = POINT(j)
			 
			 POINT(k) = 0

		 ENDDO
		!***

		 F(i) = FI

	 ENDIF
	!---
	!
	!-------------------------------
	!ADDRESS SETTING & CONTRIBUTIONS
	!-------------------------------
	 P = (/IN, IS, ID/)
		 
	 DO c1 = 0, CORR(ii)-1
		 DO c2 = 0, CORR(ii)-1

			 DO jj = 1, 3
				 DO kk = 1, 3

					 IF (P(jj).NE.0 .AND. P(kk).NE.0) THEN

						!--------------
						!ADRESS SETTING
						!--------------
						 ADDJJ = SUM(NLEV(1:ii+c1-1)) + P(jj)
						 ADDKK = SUM(NLEV(1:ii+c2-1)) + P(kk)
						!---
						!
						!------------------------------
						!FULL IJA-STORE FOR GMAT MATRIX
						!------------------------------
						 IF(c1 .EQ. 0 .AND. c2 .EQ. 0) THEN

							 IF(ADDJJ.NE.0 .AND. ADDKK.NE.0) THEN 

								 GMAT%VALL1 = WORK(jj)*WORK(kk)*(1/D(i)) 

								 CALL SBSIJ(GMAT%N1,GMAT%IA1,GMAT%JA1,GMAT%A1,GMAT%IWK1,P(jj),P(kk),&
											GMAT%VALL1,MODE,SWITCH,NOUTT,IERR)

								!AN ERROR HAS OCCURRED. STOP PROCESS.
								 IF (IERR==701 .OR. IERR==702) STOP

							 ENDIF

						 ENDIF
						!--- 	
						!
						!--------------------------------------------------------------
						!ENTRIES CALCULATION: ONLY COMPUTE UPPER DIAGONAL CONTRIBUTIONS
						!--------------------------------------------------------------
						 IF(ADDJJ.LE.ADDKK) THEN 

							 VALL = INVG(ii,1+c1,1+c2)*WORK(jj)*WORK(kk)*(1/D(i)) 

							 CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,ADDJJ,ADDKK,VALL,MODE,SWITCH,NOUTT,IERR)

							!AN ERROR HAS OCCURRED. STOP PROCESS.
							 IF (IERR==701 .OR. IERR==702) STOP 

						 ENDIF
						!---

					 ENDIF

				 ENDDO
			 ENDDO

		 ENDDO
	 ENDDO
	!---

 ENDDO                                               
!---
!
!------------
!ERROR ADVICE
!------------
 IF(PFLAG==1) THEN
	 PRINT*, ''
	 PRINT*, 'BEWARE! PEDIGREE INCONSISTENCIES FOUND.'
	 PRINT*, 'SOME IDs HAS BEEN SET TO 0.'
	 PRINT*, 'CHECK ERROR FILE.'
 ELSE
 	 WRITE(3,*) ''
 	 WRITE(3,'(a35)') 'NO INCONSISTENCIES HAVE BEEN FOUND.'
 ENDIF
!---
!
!--------------------
!CLOSE OPRATIVE FILES
!--------------------
 CLOSE (UNIT=2)
 CLOSE (UNIT=3)
!
 END SUBROUTINE	INVA
!***
!
!*****************************************************************************************
!INTERNAL SUBROUTINE GIBBS--------------------------------------------- BY SML. 01/06/2010
!*****************************************************************************************
!
!WITHIN EACH ROUND, THIS ROUTINE SAMPLES FROM THE TRANSITION KERNEL OF A MARKOV CHAIN WITH
!THE MODEL'S PARAMETER VECTOR JOINT POSTERIOR DENSITY AS THE STATIONARY DISTRIBUTION, USI-
!NG A SINGLE-SITE SYSTEMATIC GIBBS SAMPLER ALGORITHM. THE TRANSITION KERNEL IS THE PRODUCT
!OF ALL FULL CONDITIONAL POSTERIORS DENSITIES, GIVEN THEY ARE KNOWN UP TO PROPORTIONALITY. 
!IT UPDATES THE WHOLE PARAMETER VECTOR SAMPLING FROM THE APPROPIATE DENSITIES.
!***
!
!***VERSION 2.0***
!MATERNAL ANIMAL MODEL WITH CORRELATED RESIDUALS. REQUIRE FULL IJA-STORE FOR THE NUMERATOR
!RELATIONSHIP AND WW MATRICES. USE A GRIDDY-GIBBS ALGORITHM TO SAMPLE FROM THE CORRELATION
!PARAMETER FULL CONDITIONAL DENSITY.
!
!ALGORITHM DESCRIPTION********************************************************************
!THEORY BASED ON SORENSEN & GIANOLA BOOK (2002). CODE ADPATED FROM "GS_FCG D" F77 PROGRAM,
!WRITTEN BY PROFESSOR R.J.C. CANTET, AND FROM "GIBBS1.F90" PROGRAM, WRITTEN BY PROFESSOR I.
!MISTZAL. GRIDDY GIBBS ALGORITHM AS DESCRIBED BY...
!UPDATING ORDER: S2ER, TITA, pPAR, G-MATRICES. REQUIERE	3 WITHIN-BLOCK LOOPS.
!*****************************************************************************************
!
!ATT. THIS SYMBOL MEANS "ATTENTION: GIVEN THE DATA FILE".
!
!ATT. BEWARE! THIS SUBROUTINE MUST BE COMPILED WITHIN "MAM_GS v1.0" PROGRAM.
!***
!
 SUBROUTINE GIBBS(ROUND)
!
!INTENT(IN): ROUND - ITERATION NUMBER.
!***
!
!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE
!
!---------------
!DUMMY ARGUMENTS
!---------------
 INTEGER(K4B), INTENT(IN):: ROUND
!
!---------------
!LOCAL VARIABLES 
!---------------
!LOCATION PARAMETERS SAMPLING
 INTEGER(K4B):: RNDP(NEFF)
 REAL(DP):: DIAG
 REAL(DP):: TEMPOR(NEQ)
 INTEGER(K4B), ALLOCATABLE:: OFFSET(:,:,:)
 REAL(DP), ALLOCATABLE:: G_BLOCK(:,:)

!ERROR VARIANCE SAMPLING
 INTEGER(K4B):: COUNTER
 REAL(DP):: VE2, S2E, SSR, NEWR

!CORRELATION PARAMETER SAMPLING
 INTEGER(K4B):: IDX, INDI
 REAL(DP):: SSR1(NGRID), DET1(NGRID), DETR1(NGRID), CDF(NGRID)
 REAL(DP):: MAXV, WJ, WJ1, UNI

!GENETIC VARIANCE COMPONENTS SAMPLING
 INTEGER(K4B):: M1, M2, N1, N2, G1, G2
 REAL(DP):: QFORM(NEFF, MAXX, MAXX)
 INTEGER(K4B):: RANK
 REAL(DP), ALLOCATABLE:: NEWG(:)

!GIW VARIABLES
 REAL(DP), SAVE:: VP0		!CONDITIONAL POSTERIOR DIRECT BV VARIANCE DFs.
 REAL(DP), SAVE:: VP1		!CONDITIONAL POSTERIOR ADJUSTED MATERNAL BV VARIANCE DFs.
 REAL(DP), SAVE:: S0(2,2)	!PRIOR GENETIC COVARIANCE SCALE MATRIX.
 REAL(DP):: S2S, S2L, TAUP	!CONDITIONAL POSTERIOR PARAMETERS.		 
 REAL(DP):: TAU, LAMBDA			 
 
!DUMMY VARIABLES.
 INTEGER(K4B):: i, j, l, m, ii, jj, kk, t
!---
!
!*********
!PROCEDURE
!*********
!
!---------------------------
!PRINT PARAMETER MCMC CHAINS
!---------------------------
!ATT. BEWARE OUTPUT FORMATS.
 WRITE(25,'(100(f11.5,2x))') G(6,1,1), pPAR, G(5,1,1), G(7,1,1), G(7,2,1), G(7,2,2)
!---
!
!--------------------------
!INITIALIZE LOCAL VARIABLES
!--------------------------
!LOCATION PARAMETER SAMPLING
 COUNTER = 0
       t = 0

!ERROR VARIANCE SAMPLING
 SSR = 0.D0
!
!CORRELATION PARAMETER SAMPLING
!ATT. RESET GLOBAL GRID
 GRIDD(1) = MIN_VAL 
 DO i = 2, NGRID
     GRIDD(i) = GRIDD(i-1) + INCR
 ENDDO    
!---
!
!----------------------------
!LOCATION PARAMETERS SAMPLING
!----------------------------
 TEMPOR = WY !INITIALIZE "TEMPOR" WITH STORED WY VECTOR.

!SET OFFSETS AND POINTERS FOR COVARIANCE MATRICES. 
 CALL SET_RANDOM_BLOCKS (RNDP,OFFSET,G_BLOCK)
!
!-------------------
!CYCLE ALONG EFFECTS
!-------------------
 DO ii = 1, NEFF 
      DO jj = 1, NLEV1(ii)

         i = SUM(NLEV1(1:ii-1)) + jj

		 DIAG = 0

		!CONTRIBUTIONS DUE TO WW MATRIX:
		 DO j = WW%IA1(i), WW%IA1(i+1)-1
			 IF (i .NE. WW%JA1(j)) THEN
				 TEMPOR(i) =  TEMPOR(i) - WW%A1(j)*TITA(WW%JA1(j))
			 ELSE
				 DIAG = WW%A1(j)
			 ENDIF
		 ENDDO

		!CONTRIBUTIONS DUE TO INVERSE COVARIANCE MATRICES:
		 SELECT CASE(TYPO(ii))
	
 			!IF IT IS A FIXED EFFECT:
			 CASE (FIXX)

 				!ADD A VERY SMALL QUANTITY TO THE DIAGONAL ENTRY TIMES UPDATED R:
				 DIAG = DIAG + (1.D-7)

			!IF IT IS A RANDOM iid EFFECT:
			 CASE (RND1)
			
				 DIAG = DIAG + INVG(ii,1,1)

			!IF IT IS A RANDOM ANIMAL EFFECT:
			 CASE (RND2)

				 l = RNDP(ii)

				 IF (l .NE. 0) THEN

					 DO m = 1, CORR(l)
		 				 DO j = GMAT%IA1(i-OFFSET(ii,m,1)),&
							    GMAT%IA1(i-OFFSET(ii,m,1)+1)-1

							 IF (i .NE. GMAT%JA1(j) + OFFSET(ii,m,2)) THEN

								 TEMPOR(i) =  TEMPOR(i) - GMAT%A1(j)*&
											  G_BLOCK(ii,m)*TITA(GMAT%JA1(j)+&
											  OFFSET(ii,m,2))
							 ELSE
								 DIAG = DIAG + G_BLOCK(ii,m)*GMAT%A1(j)
					
							 ENDIF

						 ENDDO
			         ENDDO

				 ENDIF    

			!IF IT IS A RANDOM EFFECT WITH SPECIFIC COVARIANCE STRUCTURE:
			 CASE (RND3)

				!SET BLOCKS OFFSET
				 IF (COUNTER .EQ. 0) THEN
					      kk = 1
					       t = t+1
					 COUNTER = BK(t)%FK-1 !RESET COUNTER IF INDIVIDUAL BELONGS TO A NEW BLOCK
					   START = i
				 ELSE
					 COUNTER = COUNTER-1
					      kk = kk+1
				 ENDIF
				!---
				!
				!COMPUTE CONTRIBUTIONS
				 IF(t .NE. NBKS) THEN
					 DO j = 1, BK(t)%FK 
						 IF (i .NE. (START+j-1)) THEN
							 TEMPOR(i) =  TEMPOR(i) - BK(t)%RINV(kk,j)*INVG(ii,1,1)*TITA(START+j-1)
						 ELSE
							 DIAG = DIAG + BK(t)%RINV(kk,kk)*INVG(ii,1,1)
						 ENDIF
					 ENDDO
				 ELSE
					 DIAG = DIAG + BK(t)%RINV(kk,kk)*INVG(ii,1,1)
				 ENDIF
				!---

			!ELSE:
			 CASE DEFAULT
				 PRINT*, 'UNIMPLEMENTED RANDOM EFFECT', TYPO(ii)
				 STOP "FATAL ERROR: PROGRAM STOPPED"
		
		 END SELECT
		!***

		!COMPUTE POSTERIOR MEAN
		 TITA(i) = TEMPOR(i)/DIAG

		!SAMPLE TITA FROM A GAUSSIAN DENSITY
		 TITA(i) = TITA(i) + (1.D0 * GASDEV(SEED)) / DSQRT(DIAG)

	 ENDDO
 ENDDO
!---
!
!------------------------------
!CORRELATION PARAMETER SAMPLING
!------------------------------
!
!-------------------------------------
!LOOP 1 - SEARCH FOR LOCAL GRID VALUES
!-------------------------------------
 SSR1 = 0.D0; DET1 = 0.D0; DETR1 = 0.D0;

 START = SUM(NLEV1(1:NEFF-3))

!CYCLE AMONG BLOCKS
 DO BLK = 1, NBKS

	!------------------------------
	!CYCLE AMONG GLOBAL GRID VALUES
	!------------------------------
	 DO m = 1, NGRID

	 	!RESET WORKING VARIABLES
		 BK(BLK)%R = 0; BK(BLK)%RINV = 0

		!COMPUTE Rk-MATRIX FOR EACH pPAR (GLOBAL GRID)
		 DO i = 1, BK(BLK)%FK
			 BK(BLK)%R(i,i) = 1.D0
		 ENDDO
		 BK(BLK)%R = BK(BLK)%R + BK(BLK)%S*GRIDD(m)

		!INVERT Rk-MATRIX FOR EACH pPAR (GLOBAL GRID)
 		 IF(BLK.NE.NBKS) THEN
			 BK(BLK)%RINV = BK(BLK)%R
			 CALL GINV(BK(BLK)%RINV,BK(BLK)%FK,BK(BLK)%FK,1.D-12,RANK,DET1(m))
				 IF(RANK .LT. BK(BLK)%FK) THEN
					 PRINT*, ''
					 PRINT*, '-----------'
					 WRITE(*,'(a30,i3,a8,i1,a27)') 'FATAL ERROR: RINV FOR FAMILY ', BLK,&
												   ', GRIDD ', m,', IS NOT POSITIVE DEFINITE.' 
					 PRINT*, '-----------'
					 PRINT*, ''
					  STOP
				 ENDIF
			 DET1(m) = LOG(DET1(m))
		 ELSE
			 DO i = 1, BK(BLK)%FK	
				 BK(BLK)%RINV(i,i) = BK(BLK)%R(i,i)
			 ENDDO
		 ENDIF	

		!COMPUTE QFORM r'RINVr FOR EACH pPAR (GLOBAL GRID)
		 DO i = 1, BK(BLK)%FK
			 SSR1(m) = SSR1(m) + TITA(START+i)*TITA(START+i)*BK(BLK)%RINV(i,i)
			 IF(BLK.NE.NBKS) THEN
				 DO j = i+1, BK(BLK)%FK
					 SSR1(m) = SSR1(m) + 2*TITA(START+i)*TITA(START+j)*BK(BLK)%RINV(i,j)
				 ENDDO
			 ENDIF
		 ENDDO

		!SUM UP R-MATRIX LOG-DETERMINANTS FOR EACH pPAR (GLOBAL GRID)
		 IF(BLK.EQ.NBKS) DET1(m) = 0.D0
		 DETR1(m) = DETR1(m) + DET1(m)

	 ENDDO
	 START = START + BK(BLK)%FK
	!---

 ENDDO
!----

!EVALUATE LOG-FUNCTION
 DO j = 1, NGRID
	 SSR1(j) = -0.5D0*((SSR1(j)*INVG(6,1,1)) + DETR1(j))
 ENDDO
!---

!SEARCH MAX{LOG-FUNCTION}
 IDX = 0
 DO j = 1, NGRID
	 IF(SSR1(j) == MAXVAL(SSR1)) THEN
		 IDX=j
		 EXIT
	 ENDIF
 ENDDO
 MAXV = GRIDD(IDX)
!---
!
!RESET LOCAL GRIDD
 DO j = 1, NGRID
      GRIDD(j) = MAXV+((j-MID_VAL)*TUN)
 ENDDO
! 
!RESET WORKING VARIABLES
 SSR1 = 0.D0; DET1 = 0.D0; DETR1 = 0.D0;
!
!------------------------------------------------------------
!LOOP 2 - COMPUTE RESIDUALS & LOG-DETERMINANTS FOR LOCAL GRID
!------------------------------------------------------------
 START = SUM(NLEV1(1:NEFF-3))
!
!CYCLE AMONG BLOCKS
 DO BLK = 1, NBKS

	!-----------------------------
	!CYCLE AMONG LOCAL GRID VALUES
	!-----------------------------
	 DO m = 1, NGRID

	 	!RESET WORKING VARIABLES
		 BK(BLK)%R = 0; BK(BLK)%RINV = 0

		!COMPUTE Rk-MATRIX FOR EACH pPAR (GLOBAL GRID)
		 DO i = 1, BK(BLK)%FK
			 BK(BLK)%R(i,i) = 1.D0
		 ENDDO
		 BK(BLK)%R = BK(BLK)%R + BK(BLK)%S*GRIDD(m)

		!INVERT Rk-MATRIX FOR EACH pPAR (GLOBAL GRID)
 		 IF(BLK.NE.NBKS) THEN
			 BK(BLK)%RINV = BK(BLK)%R
			 CALL GINV(BK(BLK)%RINV,BK(BLK)%FK,BK(BLK)%FK,1.D-12,RANK,DET1(m))
				 IF(RANK .LT. BK(BLK)%FK) THEN
					 PRINT*, ''
					 PRINT*, '-----------'
					 WRITE(*,'(a30,i3,a8,i1,a27)') 'FATAL ERROR: RINV FOR FAMILY ', BLK,&
												   ', GRIDD ', m,', IS NOT POSITIVE DEFINITE.' 
					 PRINT*, '-----------'
					 PRINT*, ''
					  STOP
				 ENDIF
			 DET1(m) = LOG(DET1(m))
		 ELSE
			 DO i = 1, BK(BLK)%FK	
				 BK(BLK)%RINV(i,i) = BK(BLK)%R(i,i)
			 ENDDO
		 ENDIF	

		!COMPUTE QFORM r'RINVr FOR EACH pPAR (GLOBAL GRID)
		 DO i = 1, BK(BLK)%FK
			 SSR1(m) = SSR1(m) + TITA(START+i)*TITA(START+i)*BK(BLK)%RINV(i,i)
			 IF(BLK.NE.NBKS) THEN
				 DO j = i+1, BK(BLK)%FK
					 SSR1(m) = SSR1(m) + 2*TITA(START+i)*TITA(START+j)*BK(BLK)%RINV(i,j)
				 ENDDO
			 ENDIF
		 ENDDO

		!SUM UP R-MATRIX LOG-DETERMINANTS FOR EACH pPAR (GLOBAL GRID)
		 IF(BLK.EQ.NBKS) DET1(m) = 0.D0
		 DETR1(m) = DETR1(m) + DET1(m)

	 ENDDO
	 START = START + BK(BLK)%FK
	!---

 ENDDO
!---
!
!-------------
!pPAR SAMPLING
!-------------
!FOLLOWING ALGORITHM BY LIESENFELD, R. (2008): FINANCIAL DATA ANALYSIS, AND DR. LUC BAUWENS ADVICE.
!LOGARITHM OF THE SUM IS FOLLOWING THE ALGORITHM BY PRIMEAUX (2005).
 WJ1=0.D0
  WJ=0.D0

!STEP 1. COMPUTE w(j) FROM THE p-PARAMETER FULL CONDITIONAL DENSITY KERNEL
 DO j = 1, NGRID

	 SSR1(j) = -0.5D0*((SSR1(j)*INVG(6,1,1)) + DETR1(j))

	!ACCUMULATOR
	 IF (j==1) THEN
		 WJ = SSR1(j)
		 CYCLE
	 ENDIF
	 WJ1 = SSR1(j) - WJ 
	 WJ1 = EXP(WJ1) + 1.D0
	 WJ1 = LOG(WJ1)
	  WJ = WJ1 + WJ

 ENDDO
!---
!
!STEP 2. APPROXIMATE p-PARAMETER FULL CONDITIONAL CDF
 CDF = 0.D0
 DO j = 1, NGRID
	 DO t = 1, j
		 CDF(j) = CDF(j) + EXP(SSR1(t) - WJ)
	 ENDDO
 ENDDO
!---
!
!STEP 3. SAMPLE A UNIFORM[MIN(CDF), MAX(CDF)] DEVIATE
 INDI = 0
 t=0 !COUNTS THE NUMBER OF RANDOM UNIFORM DEVIATES SAMPLED
 DO WHILE (INDI == 0)
	 t=t+1
	 UNI = RAN(SEED)
	 IF(UNI .GE. CDF(1) .AND. UNI .LE. CDF(NGRID)) INDI = 1

	!CHECK FOR A LIMITED NUMBER OF SAMPLES
	 IF(t==20) THEN
		 PRINT*, ''
		 PRINT*, 'BEWARE! TRUNCATED CDF AT CYCLE: ', ROUND
		 PRINT*, 'CDF BETWEEN ', CDF(1), ' AND ', CDF(NGRID)
		 PRINT*, ''
	 ENDIF
	 IF(t==1000) STOP "FATAL ERROR: PROGRAM TERMINATED"
 
 ENDDO
!---
!
!STEP 4. SEARCH FOR INDEX 'IDX' SUCH THAT CDF(IDX) < UNI < CDF(IDX+1)
 IDX = 0
 DO j = 1, NGRID
	 IF(CDF(j) < UNI) THEN
		 IDX=j
	 ENDIF
 ENDDO
!---
!
!STEP 5. UPDATE pPAR BY LINEAR INTERPOLATION BETWEEN GRIDD(IDX) & GRIDD(IDX+1)
 pPAR = GRIDD(IDX) + (GRIDD(IDX+1) - GRIDD(IDX)) * (UNI - CDF(IDX))/(CDF(IDX+1) - CDF(IDX))
!
!TEST FOR SAMPLE OUT OF PARAMETER SPACE
 IF(pPAR < -1.D0 .OR.  pPAR > 1.D0) STOP " BEWARE! p_PAR SAMPLE OUT OF PARAMETRIC SPACE!"
!---
!
!-----------------------
!ERROR VARIANCE SAMPLING
!-----------------------
!COMPUTE QUADRATIC FORM r'RINVr FOR UPDATED pPAR
 SSR = 0.D0	!INITIALIZE SSR

 START = SUM(NLEV1(1:NEFF-3))

!CYCLE AMONG BLOCKS
 DO BLK = 1, NBKS

 	!RESET WORKING VARIABLES
	 BK(BLK)%R = 0; BK(BLK)%RINV = 0

	!COMPUTE Rk-MATRIX FOR UPDATED pPAR
	 DO i = 1, BK(BLK)%FK
		 BK(BLK)%R(i,i) = 1.D0
	 ENDDO
	 BK(BLK)%R = BK(BLK)%R + BK(BLK)%S*pPAR

	!INVERT Rk-MATRIX FOR UPDATED pPAR
	 IF(BLK.NE.NBKS) THEN
		 BK(BLK)%RINV = BK(BLK)%R
		 CALL GINV(BK(BLK)%RINV,BK(BLK)%FK,BK(BLK)%FK,1.D-12,RANK)
			 IF(RANK .LT. BK(BLK)%FK) THEN
				 PRINT*, ''
				 PRINT*, '-----------'
				 WRITE(*,'(a30,i3,a27)') 'FATAL ERROR: RINV FOR FAMILY ', BLK,&
											   'IS NOT POSITIVE DEFINITE.' 
				 PRINT*, '-----------'
				 PRINT*, ''
				  STOP
			 ENDIF
	 ELSE
		 DO i = 1, BK(BLK)%FK	
			 BK(BLK)%RINV(i,i) = BK(BLK)%R(i,i)
		 ENDDO
	 ENDIF	

	!COMPUTE QFORM r'RINVr FOR UPDATED pPAR
	 DO i = 1, BK(BLK)%FK
		 SSR = SSR + TITA(START+i)*TITA(START+i)*BK(BLK)%RINV(i,i)
		 IF(BLK.NE.NBKS) THEN
			 DO j = i+1, BK(BLK)%FK
				 SSR = SSR + 2*TITA(START+i)*TITA(START+j)*BK(BLK)%RINV(i,j)
			 ENDDO
		 ENDIF
	 ENDDO
	 START = START + BK(BLK)%FK

 ENDDO
!---

!SET SCALED INVERTED CHI-SQUARE PARAMETERS
!DEGREES OF BELIEF 
 VE2 = 1.D0*(NRDF + VE1)

!SCALE PARAMETER 
 S2E = (SSR + (G0(6,1,1) * (VE1+2)))

!SAMPLE FROM A SCALED INVERTED CHI-SQUARE DENSITY
 NEWR = ICHSQ(VE2, S2E, SEED)

!TEST FOR BIG SAMPLE VALUE
 IF(NEWR > 20000.D0) STOP " BEWARE! R SAMPLE VALUE TOO BIG!"

!UPDATE ERROR VARIANCE
 G(6,1,1) = NEWR
!---
!
!----------------------------
!VARIANCE COMPONENTS SAMPLING
!----------------------------
 QFORM = 0.D0

!COMPUTE QUADRATIC AND BILINEAR FORMS 
 DO i = 1, NEFF

	!IF IT IS A RANDOM EFFECT, NOT CORRELATED WITH A PREVIOUS ONE:
	 IF (CORR(i) .GT. 0) THEN

		!SET j=0 FOR DIRECT EFFECT AND j=1 FOR MATERNAL
	     DO j = 0, CORR(i) - 1

			!SET k=0 FOR DIRECT EFFECT AND k=1 FOR MATERNAL
			 DO k = 0, CORR(i) - 1

				!SET ADDRESS WHERE i-EFFECT START & END.
				 M1 = SUM(NLEV1(1:(i+j)-1)) + 1; M2 = SUM(NLEV1(1:(i+j)-1)) + NLEV1(i)
				 N1 = SUM(NLEV1(1:(i+k)-1)) + 1; N2 = SUM(NLEV1(1:(i+k)-1)) + NLEV1(i)

				!SET G1=1 FOR DIRECT EFFECT AND G1=2 FOR MATERNAL
				 G1= j + 1

				!SET G2=1 FOR DIRECT EFFECT AND G2=2 FOR MATERNAL
				 G2= k + 1

				!COMPUTE ONLY BELOW DIAGONAL BILINEAR FORMS
				 IF (G1 .LE. G2) THEN

					!FOR IID RANDOM EFFECTS
					 IF(TYPO(i) .EQ. RND1) THEN

						!COMPUTE QUADRATIC FORM
						 QFORM(i,G1,G2) = DOT_PRODUCT(TITA(M1:M2),TITA(N1:N2))

					!FOR RANDOM ANIMAL EFFECTS
					 ELSEIF(TYPO(i) .EQ. RND2) THEN

						!COMPUTE QUADRATIC FORMS
						 QFORM(i,G1,G2) = QFORM_IJA(TITA(M1:M2),GMAT,TITA(N1:N2))

   						!COPY ABOVE DIAGONAL BILINEAR FORMS
						 QFORM(i,G2,G1) = QFORM(i,G1,G2)	

					 ELSEIF(TYPO(i) .EQ. RND3) THEN

						 CONTINUE !SKIP EFFECT

					 ELSE

						STOP "BEWARE! CHECK <<TYPO>> AND <<CORR>> DECLARATION"

					 ENDIF

				 ENDIF

			 ENDDO

		 ENDDO
		!***

		!SET SCALED INVERTED WISHART PARAMETERS, SAMPLE AND UPDATE G.
		 j = CORR(i)

	     IF (j .GT. 0) THEN

			!FOR IID RANDOM EFFECTS
			 IF(TYPO(i) .EQ. RND1) THEN

				!SET SCALED INVERTED CHI-SQUARE PARAMETERS
				!DEGREES OF BELIEF 
				 VE2 = 1.D0*(NLEV1(i) + VEM1)

				!SCALE PARAMETER 
 				 S2E = (QFORM(i,1,1) + (G0(i,1,1) * (VEM1+2)))

				!SAMPLE FROM A SCALED INVERTED CHI-SQUARE DENSITY
			 	 ALLOCATE(NEWG(1))

                 NEWG(1) = ICHSQ(VE2, S2E, SEED)
 
				!UPDATE IID RANDOM EFFECT VARIANCE
				 G(i,1,1) = NEWG(1)

			 	 DEALLOCATE(NEWG)


			!FOR RANDOM ANIMAL EFFECTS
			 ELSEIF(TYPO(i) .EQ. RND2) THEN

				!********************************************************************************************
				!SAMPLING THE GENETIC COVARIANCE MATRIX THROUGH A GIW APPROACH (MUNILLA & CANTET, JABG, 2012)
				!********************************************************************************************
				!THE GIW APPROACH IS BASED ON A DECOMPOSITION OF THE COVARIANCE MATRIX ACCORDINGLY
				!TO THE BARTLETT'S DECOMPOSITION. (CO)VARIANCES COMPONENTES ARE DRAWN FROM 3 DIFFE-
				!RENT DENSITIES. 

				!1. SET PRIOR PARAMETERS
				 IF(ROUND == 1) THEN

					 S0(1,1) =  G0(i,1,1)*(V0 + 2)
					 S0(1,2) =  S0(1,1)*(G0(i,1,2)/G0(i,1,1))
					 S0(2,1) =  S0(1,2)
					 S0(2,2) = (G0(i,2,2) - (G0(i,1,2)**2/G0(i,1,1)))*(V1 + 3)
					 S0(2,2) =  S0(2,2) + (S0(1,2)**2/S0(1,1))

					 VP0 = 1.D0*(NLEV1(i) + V0)
					 VP1 = 1.D0*(NLEV1(i) + V1)

				 ENDIF
				!---

				!2. COMPUTE CONDITIONAL POSTERIOR GENETIC COVARIANCE SCALE MATRIX.

					 QFORM(i,1:j,1:j) = QFORM(i,1:j,1:j) + S0(1:j,1:j)

				!3. SAMPLE S2A0 FROM A SCALED INVERTED CHI-SQUARE DENSITY

					!SCALE PARAMETER 
					 S2S = QFORM(i,1,1)

					!SAMPLE FROM A SCALED INVERTED CHI-SQUARE DENSITY
					 G(i,1,1) = ICHSQ(VP0, S2S, SEED)

				!4. SAMPLE LAMBDA FROM A SCALED INVERTED CHI-SQUARE DENSITY

					!SET SCALE PARAMETER 
					 S2L = QFORM(i,2,2)
					 S2L = S2L - (QFORM(i,1,2)**2/QFORM(i,1,1))

					!SAMPLE FROM A SCALED INVERTED CHI-SQUARE DENSITY
					 LAMBDA = ICHSQ((VP1+1), S2L, SEED)

   				!5. SAMPLE TAU|LAMBDA FROM A GAUSSIAN DENSITY

					!FIRST, SET CONDITIONAL POSTERIOR MEAN
					 TAUP =  (QFORM(i,1,2)/QFORM(i,1,1))

					!FINALLY, SAMPLE FROM A GAUSSIAN DENSITY
					 TAU = TAUP + (1.D0 * GASDEV(SEED)) * DSQRT(LAMBDA/QFORM(i,1,1))

  				!6. RETRIEVE G0 THROUGH THE BARTLETT DECOMPOSITION
					G(i,1,2) = G(i,1,1)*TAU
					G(i,2,1) = G(i,1,2)
					G(i,2,2) = LAMBDA + (G(i,1,1)*TAU*TAU)

					!FINALLY, INVERT G TO CHECK IF IT IS POSITIVE DEFINITE
					 GTEST = G
					 CALL GINV(GTEST(i,1:j,1:j),j,MAXX,1.D-12,RANK)
	
						!SECURITY DEVICE: CHECK OUT IF SAMPLED MATRIX IS POSITIVE DEFINITE.
						 IF(RANK .LT. j) THEN
							 PRINT*, ''
							 PRINT*, '-----------'
							 WRITE(*,'(a26,i2,a24)') 'FATAL ERROR: MATRIX G(', i, ') IS &
													  &NOT POSITIVE DEFINITE.' 
							 WRITE(*,'(a13,i7)') 'ROUND:', ROUND
							 PRINT*, '-----------'
							 PRINT*, ''
							  STOP
						 ENDIF

				!**********************************************************************************
				!ENDS GIW APPROACH-----------------------------------------------------------------
				!**********************************************************************************

				 ELSEIF(TYPO(i) .EQ. RND3) THEN

					 CONTINUE !SKIP EFFECT

			 ELSE

				STOP "BEWARE! CHECK <<TYPO>> AND <<CORR>> DECLARATION"

			 ENDIF

	     ENDIF
		!***

	 ENDIF

 ENDDO
!---

 END SUBROUTINE GIBBS
!***
!
!*****************************************************************************************
!INTERNAL SUBROUTINE SET_RANDOM_BLOCKS -------------------------------- BY SML. 21/05/2009
!*****************************************************************************************
!
!SET POINTERS AND VARIANCE COMPONENTS ASSOCIATED TO EACH BLOCK OF A COVARIANCE MATRIX WITH
!KRONECKER STRUCTURE, SUCH AS THE RELATIONSHIP MATRIX IN A CORRELATED TRAIT MODEL. BLOCK j 
!STANDS FOR AN EQUATION OF THE jth-CORRELATED COMPONENT IN THIS CASE.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!PROFESSOR IGNACY MISTZAL'S CODE, SLIGHTLY ADAPTED TO ONE-EQUATION BLOCKS. 
!*****************************************************************************************
!
!ATT. THIS SYMBOL MEANS "ATTENTION: GIVEN THE DATA FILE".
!***

 SUBROUTINE SET_RANDOM_BLOCKS (RNDP,OFFSET,G_BLOCK)

!INTENT(IN)::		   RNDP(NEFF) - POSITION OF THE FIRST EFFECT OF THE CORRELATED BLOCK. 
!INTENT(IN)::  G_BLOCK(NEFF,MAXX) - RETURNS THE CORRESPONDING VARIANCE COMPONENT.
!INTENT(IN):: OFFSET(NEFF,MAXX,2) - RETURNS ADDRESSES OF THE BLOCK TO BE COMPUTED.

!***********
!DECLARATION
!***********

!DUMMY ARGUMENTS
 INTEGER(K4B), INTENT(INOUT):: RNDP(:)
 INTEGER(K4B), ALLOCATABLE:: OFFSET(:,:,:)
 REAL(DP), ALLOCATABLE:: G_BLOCK(:,:)

!LOCAL VARIABLES 
 INTEGER(K4B):: i, j, k

!*********
!PROCEDURE
!*********
  
!INITIALIZATION.
 ALLOCATE(OFFSET(NEFF,MAXX,2), G_BLOCK(NEFF,MAXX))

 RNDP = 0; OFFSET = 0; G_BLOCK = 0         
  
!SETTING RNDP VECTOR
 i=1
 DO WHILE(i .LE. NEFF)
	 IF(CORR(i) .GT. 0) THEN
		 DO j = 0, CORR(i)-1
			 RNDP(i+j) = i
		 ENDDO
		 i = i + CORR(i)  
     ELSE
		 i = i + 1
     ENDIF
 ENDDO
!***

!SETTING OFFSETS & G_BLOCK
 DO i = 1, NEFF
	 k = RNDP(i)
     IF(k /= 0) THEN
		 DO j = 1, CORR(k)
			 OFFSET(i,j,1) = SUM(NLEV1(1:i-1))     !ROW POINTER.
			 OFFSET(i,j,2) = SUM(NLEV1(1:(k+j-2))) !COLUMN POINTER.
              G_BLOCK(i,j) = INVG(k,(i-k+1),j)	   !INVG MATRIX ENTRY.
		 ENDDO
	 ENDIF
 ENDDO
!***

 END SUBROUTINE SET_RANDOM_BLOCKS  
!***
!
!*****************************************************************************************
!INTERNAL SUBROUTINE SET_ADDRESS -------------------------------------- BY SML. 05/06/2009
!*****************************************************************************************
!
!SET (EFFECT, LEVEL, TRAIT) ADDRESSES FOR EACH DATA FILE LINE.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!LOOPS OVER EFFECTS AND TRAITS AND SET ADDRESSES GIVEN EFFECT TYPE. 
!*****************************************************************************************
!
!ATT. BEWARE! THIS SUBROUTINE MUST BE COMPILED WITHIN "MAM_GS v1.0" PROGRAM.
!
 SUBROUTINE SET_ADDRESS

!LOCAL VARIABLES 
 INTEGER(K4B):: i, j

!*********
!PROCEDURE
!*********

 DO i = 1, NEFF !LOOP OVER EFFECTS

	 DO j = 1, NTRT !LOOP OVER TRAITS
	
		 START = SUM(NLEV(1:i-1))*NTRT + j

		 SELECT CASE(EFFT(i))
	
			!IF THE EFFECT IS A CATEGORICAL FACTOR:
			 CASE (CLSS)

				 IF(INDATA(i)==0) THEN
				 ADRESS(i,j) = 0
				 CYCLE
				 ENDIF
				!***

				 IF(INDATA(i)<0) THEN
				 ADRESS(i,j) = START + (-INDATA(i) - 1)*NTRT
				 CVRW(i,j) = -1.D0

				 ELSE
				 ADRESS(i,j) = START + (INDATA(i) - 1)*NTRT 
				 CVRW(i,j) = 1.D0

				 ENDIF

			!IF THE EFFECT IS A COVARIATE:
			 CASE (COVR)
					
				 CVRW(i,j) = INDATA(i)	
				
				!IF THE COVARIATE IS NOT NESTED:
				 IF (NEST(i).eq.0) THEN

					 ADRESS(i,j) = START
				
				!IF THE COVARIATE IS NESTED:
				 ELSEIF (NEST(i).gt.0.AND.NEST(i).lt.NEFF) THEN
						
					 ADRESS(i,j) = START + (INDATA(NEST(i)) - 1)*NTRT
	
				 ELSE
				 PRINT*, 'ERROR IN NESTED COVARIATE. PLEASE, CHECK VARIABLE "NEST" SETTING'
				 STOP
	
				 ENDIF

			 CASE DEFAULT
			 PRINT*, 'UNIMPLEMENTED FIXED EFFECT', i
			 STOP
		
		 END SELECT

	 ENDDO

 ENDDO
!***
  
 END SUBROUTINE SET_ADDRESS  
!***

!*****************************************************************************************
!FUNCTION QFORM_IJA --------------------------------------------------- BY SML. 19/06/2009
!*****************************************************************************************
!
!IGNACYS MISTZAL'S CODE. COMPUTE A BILINEAR FORM WHEN THE ASSOCIATED MATRIX IS SQUARE, SY-
!MMETRIC AND FULL IJA-STORED.
!***
!
 FUNCTION QFORM_IJA(X,A,Y) RESULT(Q)

!***********
!DECLARATION
!***********
 IMPLICIT NONE

!DUMMY ARGUMENT
 TYPE(SPARSE_MATRIX):: A
 REAL(DP):: X(:), Y(:)

!LOCAL VARIABLES
 REAL(DP):: Q
 INTEGER(K4B):: i, j, k
!***

!*********
!PROCEDURE
!*********

!CHECK IF THE ASSOCIATED MATRIX IS SQUARE.
 IF (SIZE(X).NE.A%N1 .OR. SIZE(Y).NE.A%N1) STOP "INCOMPATIBLE DIMENSIONS FOR THE BILINEAR&
												 FORM COMPUTATION"
!INITIALIZE
 Q = 0

 DO i = 1, A%N1
 
	 DO j = A%IA1(i), A%IA1(i+1) - 1

		 k = A%JA1(j)

		 IF (i .EQ. k) Q = Q + A%A1(j)*X(i)*Y(k)

		 IF (i .LT. k) Q = Q + A%A1(j)*(X(i)*Y(k) + X(k)*Y(i))

	 ENDDO

 ENDDO
!***
 
 END FUNCTION QFORM_IJA
!***

 END PROGRAM  RDBLK_GS
!***
